/* Generated by Cython 0.25.2 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__gtfspy__routing__label
#define __PYX_HAVE_API__gtfspy__routing__label
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "gtfspy/routing/label.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple;
struct __pyx_obj_6gtfspy_7routing_5label_LabelTime;
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;
struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount;
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute;
struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric;

/* "gtfspy/routing/label.pyx":1
 * cdef class LabelTimeSimple:             # <<<<<<<<<<<<<<
 *     """
 *     LabelTime describes entries in a Profile.
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_vtab;
  double departure_time;
  double arrival_time_target;
};


/* "gtfspy/routing/label.pyx":52
 *         return LabelTimeSimple(self.departure_time, self.arrival_time_target)
 * 
 * cdef class LabelTime:             # <<<<<<<<<<<<<<
 *     """
 *     LabelTime describes entries in a Profile.
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelTime {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *__pyx_vtab;
  double departure_time;
  double arrival_time_target;
  int first_leg_is_walk;
};


/* "gtfspy/routing/label.pyx":120
 *         return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)
 * 
 * cdef class LabelTimeWithBoardingsCount:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double departure_time
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_vtab;
  double departure_time;
  double arrival_time_target;
  int n_boardings;
  int first_leg_is_walk;
};


/* "gtfspy/routing/label.pyx":235
 * 
 * 
 * cdef class LabelVehLegCount:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double departure_time
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_vtab;
  double departure_time;
  int n_boardings;
  int first_leg_is_walk;
};


/* "gtfspy/routing/label.pyx":450
 *         return None
 * 
 * cdef class LabelTimeBoardingsAndRoute:             # <<<<<<<<<<<<<<
 *     # Label with implemented added constraint for cases when two labels are tied:
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_vtab;
  double departure_time;
  double arrival_time_target;
  int n_boardings;
  int movement_duration;
  int first_leg_is_walk;
  PyObject *previous_label;
  PyObject *connection;
};


/* "gtfspy/routing/label.pyx":584
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 * cdef class LabelTimeAndRoute:             # <<<<<<<<<<<<<<
 *     # implement added constraint for cases when two labels are tied:
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_vtab;
  double departure_time;
  double arrival_time_target;
  int movement_duration;
  int first_leg_is_walk;
  PyObject *previous_label;
  PyObject *connection;
};


/* "gtfspy/routing/label.pyx":712
 * 
 * 
 * cdef class LabelGeneric:             # <<<<<<<<<<<<<<
 *     # This class is used only for the analysis stage, when the data has been stored in a database
 *     cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \
 */
struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric {
  PyObject_HEAD
  struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelGeneric *__pyx_vtab;
  int journey_id;
  int from_stop_I;
  int to_stop_I;
  int n_boardings;
  int movement_duration;
  int journey_duration;
  int in_vehicle_duration;
  int transfer_wait_duration;
  int walking_duration;
  int pre_journey_wait_fp;
  double departure_time;
  double arrival_time_target;
};



/* "gtfspy/routing/label.pyx":1
 * cdef class LabelTimeSimple:             # <<<<<<<<<<<<<<
 *     """
 *     LabelTime describes entries in a Profile.
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeSimple {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeSimple;


/* "gtfspy/routing/label.pyx":52
 *         return LabelTimeSimple(self.departure_time, self.arrival_time_target)
 * 
 * cdef class LabelTime:             # <<<<<<<<<<<<<<
 *     """
 *     LabelTime describes entries in a Profile.
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime {
  PyObject *(*_tuple_for_ordering)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*get_copy_with_specified_departure_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, double, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*get_copy_with_walk_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *__pyx_vtabptr_6gtfspy_7routing_5label_LabelTime;


/* "gtfspy/routing/label.pyx":120
 *         return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)
 * 
 * cdef class LabelTimeWithBoardingsCount:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double departure_time
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_and_n_boardings)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  PyObject *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  PyObject *(*get_copy_with_specified_departure_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, PyObject *, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *(*get_copy_with_walk_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;


/* "gtfspy/routing/label.pyx":235
 * 
 * 
 * cdef class LabelVehLegCount:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double departure_time
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelVehLegCount {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_vtabptr_6gtfspy_7routing_5label_LabelVehLegCount;


/* "gtfspy/routing/label.pyx":450
 *         return None
 * 
 * cdef class LabelTimeBoardingsAndRoute:             # <<<<<<<<<<<<<<
 *     # Label with implemented added constraint for cases when two labels are tied:
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_and_n_boardings)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  PyObject *(*get_label_with_connection_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  PyObject *(*get_copy_with_specified_departure_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, PyObject *, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *(*get_copy_with_walk_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, double, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;


/* "gtfspy/routing/label.pyx":584
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 * cdef class LabelTimeAndRoute:             # <<<<<<<<<<<<<<
 *     # implement added constraint for cases when two labels are tied:
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeAndRoute {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_and_n_boardings)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  PyObject *(*get_label_with_connection_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  PyObject *(*get_copy_with_specified_departure_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, PyObject *, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *(*get_copy_with_walk_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, double, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeAndRoute;


/* "gtfspy/routing/label.pyx":712
 * 
 * 
 * cdef class LabelGeneric:             # <<<<<<<<<<<<<<
 *     # This class is used only for the analysis stage, when the data has been stored in a database
 *     cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \
 */

struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelGeneric {
  int (*dominates)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_finalization)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  int (*dominates_ignoring_dep_time_and_n_boardings)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  PyObject *(*get_label_with_connection_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_copy)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  PyObject *(*get_copy_with_specified_departure_time)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, PyObject *, int __pyx_skip_dispatch);
  double (*duration)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch);
  struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *(*get_copy_with_walk_added)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, double, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelGeneric *__pyx_vtabptr_6gtfspy_7routing_5label_LabelGeneric;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?\
 PyFloat_AS_DOUBLE(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* GetNameInClass.proto */
static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_9LabelTime__tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_9LabelTime_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_walk_duration, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_walk_duration, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch); /* proto*/
static double __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'gtfspy.routing.label' */
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelTime = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute = 0;
static PyTypeObject *__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric = 0;
#define __Pyx_MODULE_NAME "gtfspy.routing.label"
int __pyx_module_is_main_gtfspy__routing__label = 0;

/* Implementation of 'gtfspy.routing.label' */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_all;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_any;
static PyObject *__pyx_builtin_reversed;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_min;
static const char __pyx_k_all[] = "all";
static const char __pyx_k_any[] = "any";
static const char __pyx_k_inf[] = "inf";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_label[] = "label";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_labels[] = "labels";
static const char __pyx_k_other_i[] = "other_i";
static const char __pyx_k_duration[] = "duration";
static const char __pyx_k_get_copy[] = "get_copy";
static const char __pyx_k_new_best[] = "new_best";
static const char __pyx_k_reversed[] = "reversed";
static const char __pyx_k_survived[] = "survived";
static const char __pyx_k_candidate[] = "candidate";
static const char __pyx_k_dominated[] = "dominated";
static const char __pyx_k_dominates[] = "dominates";
static const char __pyx_k_dominator[] = "dominator";
static const char __pyx_k_new_label[] = "new_label";
static const char __pyx_k_remaining[] = "remaining";
static const char __pyx_k_to_stop_I[] = "to_stop_I";
static const char __pyx_k_best_label[] = "best_label";
static const char __pyx_k_candidates[] = "candidates";
static const char __pyx_k_connection[] = "connection";
static const char __pyx_k_journey_id[] = "journey_id";
static const char __pyx_k_label_list[] = "label_list";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_from_stop_I[] = "from_stop_I";
static const char __pyx_k_n_boardings[] = "n_boardings";
static const char __pyx_k_finalization[] = "finalization";
static const char __pyx_k_is_dominated[] = "is_dominated";
static const char __pyx_k_journey_dict[] = "journey_dict";
static const char __pyx_k_labels_other[] = "labels_other";
static const char __pyx_k_pareto_front[] = "pareto_front";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_new_remaining[] = "new_remaining";
static const char __pyx_k_survivor_list[] = "survivor_list";
static const char __pyx_k_walk_duration[] = "walk_duration";
static const char __pyx_k_departure_time[] = "departure_time";
static const char __pyx_k_previous_label[] = "previous_label";
static const char __pyx_k_n_vehicle_trips[] = "n_vehicle_trips";
static const char __pyx_k_old_partial_best[] = "old_partial_best";
static const char __pyx_k_direct_walk_label[] = "direct_walk_label";
static const char __pyx_k_first_leg_is_walk[] = "first_leg_is_walk";
static const char __pyx_k_movement_duration[] = "movement_duration";
static const char __pyx_k_ignore_n_boardings[] = "ignore_n_boardings";
static const char __pyx_k_tuple_for_ordering[] = "_tuple_for_ordering";
static const char __pyx_k_arrival_time_target[] = "arrival_time_target";
static const char __pyx_k_min_n_vehicle_trips[] = "min_n_vehicle_trips";
static const char __pyx_k_possible_dominators[] = "possible_dominators";
static const char __pyx_k_compute_pareto_front[] = "compute_pareto_front";
static const char __pyx_k_gtfspy_routing_label[] = "gtfspy.routing.label";
static const char __pyx_k_candidate_is_dominated[] = "candidate_is_dominated";
static const char __pyx_k_merge_pareto_frontiers[] = "merge_pareto_frontiers";
static const char __pyx_k_dominates_ignoring_time[] = "dominates_ignoring_time";
static const char __pyx_k_min_arrival_time_target[] = "min_arrival_time_target";
static const char __pyx_k_get_copy_with_walk_added[] = "get_copy_with_walk_added";
static const char __pyx_k_get_non_dominated_entries[] = "_get_non_dominated_entries";
static const char __pyx_k_compute_pareto_front_naive[] = "compute_pareto_front_naive";
static const char __pyx_k_compute_pareto_front_smart[] = "compute_pareto_front_smart";
static const char __pyx_k_dominates_ignoring_dep_time[] = "dominates_ignoring_dep_time";
static const char __pyx_k_current_best_labels_wo_deptime[] = "current_best_labels_wo_deptime";
static const char __pyx_k_Users_rmkujala_work_gtfspy_gtfs[] = "/Users/rmkujala/work/gtfspy/gtfspy/routing/label.pyx";
static const char __pyx_k_dominates_ignoring_dep_time_and[] = "dominates_ignoring_dep_time_and_n_boardings";
static const char __pyx_k_get_label_with_connection_added[] = "get_label_with_connection_added";
static const char __pyx_k_dominates_ignoring_dep_time_fina[] = "dominates_ignoring_dep_time_finalization";
static const char __pyx_k_get_copy_with_specified_departur[] = "get_copy_with_specified_departure_time";
static const char __pyx_k_merge_pareto_frontiers_locals__g[] = "merge_pareto_frontiers.<locals>._get_non_dominated_entries";
static const char __pyx_k_min_arrival_time_target_locals_l[] = "min_arrival_time_target.<locals>.<lambda>";
static const char __pyx_k_min_n_vehicle_trips_locals_lambd[] = "min_n_vehicle_trips.<locals>.<lambda>";
static PyObject *__pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_any;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arrival_time_target;
static PyObject *__pyx_n_s_best_label;
static PyObject *__pyx_n_s_candidate;
static PyObject *__pyx_n_s_candidate_is_dominated;
static PyObject *__pyx_n_s_candidates;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_compute_pareto_front;
static PyObject *__pyx_n_s_compute_pareto_front_naive;
static PyObject *__pyx_n_s_compute_pareto_front_smart;
static PyObject *__pyx_n_s_connection;
static PyObject *__pyx_n_s_current_best_labels_wo_deptime;
static PyObject *__pyx_n_s_departure_time;
static PyObject *__pyx_n_s_direct_walk_label;
static PyObject *__pyx_n_s_dominated;
static PyObject *__pyx_n_s_dominates;
static PyObject *__pyx_n_s_dominates_ignoring_dep_time;
static PyObject *__pyx_n_s_dominates_ignoring_dep_time_and;
static PyObject *__pyx_n_s_dominates_ignoring_dep_time_fina;
static PyObject *__pyx_n_s_dominates_ignoring_time;
static PyObject *__pyx_n_s_dominator;
static PyObject *__pyx_n_s_duration;
static PyObject *__pyx_n_s_finalization;
static PyObject *__pyx_n_s_first_leg_is_walk;
static PyObject *__pyx_n_s_from_stop_I;
static PyObject *__pyx_n_s_get_copy;
static PyObject *__pyx_n_s_get_copy_with_specified_departur;
static PyObject *__pyx_n_s_get_copy_with_walk_added;
static PyObject *__pyx_n_s_get_label_with_connection_added;
static PyObject *__pyx_n_s_get_non_dominated_entries;
static PyObject *__pyx_n_s_gtfspy_routing_label;
static PyObject *__pyx_n_s_ignore_n_boardings;
static PyObject *__pyx_n_s_inf;
static PyObject *__pyx_n_s_is_dominated;
static PyObject *__pyx_n_s_journey_dict;
static PyObject *__pyx_n_s_journey_id;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_label;
static PyObject *__pyx_n_s_label_list;
static PyObject *__pyx_n_s_labels;
static PyObject *__pyx_n_s_labels_other;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_merge_pareto_frontiers;
static PyObject *__pyx_n_s_merge_pareto_frontiers_locals__g;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_min_arrival_time_target;
static PyObject *__pyx_n_s_min_arrival_time_target_locals_l;
static PyObject *__pyx_n_s_min_n_vehicle_trips;
static PyObject *__pyx_n_s_min_n_vehicle_trips_locals_lambd;
static PyObject *__pyx_n_s_movement_duration;
static PyObject *__pyx_n_s_n_boardings;
static PyObject *__pyx_n_s_n_vehicle_trips;
static PyObject *__pyx_n_s_new_best;
static PyObject *__pyx_n_s_new_label;
static PyObject *__pyx_n_s_new_remaining;
static PyObject *__pyx_n_s_old_partial_best;
static PyObject *__pyx_n_s_other;
static PyObject *__pyx_n_s_other_i;
static PyObject *__pyx_n_s_pareto_front;
static PyObject *__pyx_n_s_possible_dominators;
static PyObject *__pyx_n_s_previous_label;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_remaining;
static PyObject *__pyx_n_s_reversed;
static PyObject *__pyx_n_s_staticmethod;
static PyObject *__pyx_n_s_survived;
static PyObject *__pyx_n_s_survivor_list;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_to_stop_I;
static PyObject *__pyx_n_s_tuple_for_ordering;
static PyObject *__pyx_n_s_walk_duration;
static PyObject *__pyx_n_s_zip;
static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_2__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_4__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_6__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_8dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_10dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_12direct_walk_label(double __pyx_v_departure_time, double __pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_first_leg_is_walk, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_2_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_4__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_6__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_8__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_12dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_14dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_18get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_20direct_walk_label(double __pyx_v_departure_time, double __pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_22duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_24get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_n_boardings, int __pyx_v_first_leg_is_walk); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_20get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_22get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_24duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_26direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_28get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_30__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_n_boardings, double __pyx_v_departure_time, PyObject *__pyx_v_first_leg_is_walk, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_12dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_18get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_20get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_22direct_walk_label(PyObject *__pyx_v_departure_time, CYTHON_UNUSED PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_compute_pareto_front_smart(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_2compute_pareto_front(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list, PyObject *__pyx_v_finalization, PyObject *__pyx_v_ignore_n_boardings); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_4compute_pareto_front_naive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_22merge_pareto_frontiers__get_non_dominated_entries(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_candidates, PyObject *__pyx_v_possible_dominators, PyObject *__pyx_v_survivor_list); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_6merge_pareto_frontiers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_labels, PyObject *__pyx_v_labels_other); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_8min_arrival_time_target(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_10min_n_vehicle_trips(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_n_boardings, int __pyx_v_movement_duration, int __pyx_v_first_leg_is_walk, PyObject *__pyx_v_connection, PyObject *__pyx_v_previous_label); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_movement_duration, int __pyx_v_first_leg_is_walk, PyObject *__pyx_v_connection, PyObject *__pyx_v_previous_label); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_journey_dict); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_state); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_departure_time); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self); /* proto */
static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeSimple(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTime(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelVehLegCount(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeAndRoute(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelGeneric(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_neg_1;
static double __pyx_k_;
static double __pyx_k__2;
static double __pyx_k__3;
static double __pyx_k__4;
static double __pyx_k__5;
static PyObject *__pyx_slice__9;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_slice__10;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_slice__12;
static PyObject *__pyx_slice__13;
static PyObject *__pyx_slice__14;
static PyObject *__pyx_slice__15;
static PyObject *__pyx_slice__16;
static PyObject *__pyx_slice__17;
static PyObject *__pyx_slice__18;
static PyObject *__pyx_slice__19;
static PyObject *__pyx_slice__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;

/* "gtfspy/routing/label.pyx":9
 *         public double arrival_time_target
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_arrival_time_target;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_arrival_time_target,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrival_time_target);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 9, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 9, __pyx_L3_error)
    } else {
      __pyx_v_departure_time = __pyx_k_;
    }
    if (values[1]) {
      __pyx_v_arrival_time_target = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_arrival_time_target == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 9, __pyx_L3_error)
    } else {
      __pyx_v_arrival_time_target = __pyx_k__2;
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 9, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), __pyx_v_departure_time, __pyx_v_arrival_time_target);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":10
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.arrival_time_target = arrival_time_target
 * 
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":11
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):
 */
  __pyx_v_self->arrival_time_target = __pyx_v_arrival_time_target;

  /* "gtfspy/routing/label.pyx":9
 *         public double arrival_time_target
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":13
 *         self.arrival_time_target = arrival_time_target
 * 
 *     def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self.departure_time, -self.arrival_time_target
 *         other_tuple = other.departure_time, -other.arrival_time_target
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_3__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple, 1, "self", 0))) __PYX_ERR(0, 13, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple, 1, "other", 0))) __PYX_ERR(0, 13, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_2__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_2__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":14
 * 
 *     def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):
 *         self_tuple = self.departure_time, -self.arrival_time_target             # <<<<<<<<<<<<<<
 *         other_tuple = other.departure_time, -other.arrival_time_target
 *         if op == 2:  # ==
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":15
 *     def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):
 *         self_tuple = self.departure_time, -self.arrival_time_target
 *         other_tuple = other.departure_time, -other.arrival_time_target             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_other->departure_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":16
 *         self_tuple = self.departure_time, -self.arrival_time_target
 *         other_tuple = other.departure_time, -other.arrival_time_target
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":17
 *         other_tuple = other.departure_time, -other.arrival_time_target
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 17, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":16
 *         self_tuple = self.departure_time, -self.arrival_time_target
 *         other_tuple = other.departure_time, -other.arrival_time_target
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":18
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":19
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 19, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":18
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":20
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":21
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":20
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":22
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":23
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":22
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":24
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":25
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":24
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":26
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":27
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     # __getstate__ and __setstate__ : for pickling
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":26
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":13
 *         self.arrival_time_target = arrival_time_target
 * 
 *     def __richcmp__(LabelTimeSimple self, LabelTimeSimple other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self.departure_time, -self.arrival_time_target
 *         other_tuple = other.departure_time, -other.arrival_time_target
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":30
 * 
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_5__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_4__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_4__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":31
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":30
 * 
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":33
 *         return self.departure_time, self.arrival_time_target
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_7__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_6__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_6__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  double __pyx_t_5;
  double __pyx_t_6;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":34
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target = state             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelTimeSimple other) except *:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 34, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_3 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_4(__pyx_t_3); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_4(__pyx_t_3), 2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
    __pyx_t_4 = NULL;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 34, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->departure_time = __pyx_t_5;
  __pyx_v_self->arrival_time_target = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":33
 *         return self.departure_time, self.arrival_time_target
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":36
 *         self.departure_time, self.arrival_time_target = state
 * 
 *     cpdef int dominates(self, LabelTimeSimple other) except *:             # <<<<<<<<<<<<<<
 *         return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_9dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_9dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 36, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":37
 * 
 *     cpdef int dominates(self, LabelTimeSimple other) except *:
 *         return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeSimple other):
 */
  __pyx_t_7 = (__pyx_v_self->departure_time >= __pyx_v_other->departure_time);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":36
 *         self.departure_time, self.arrival_time_target = state
 * 
 *     cpdef int dominates(self, LabelTimeSimple other) except *:             # <<<<<<<<<<<<<<
 *         return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_9dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_9dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple, 1, "other", 0))) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_8dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_8dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates(__pyx_v_self, __pyx_v_other, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":39
 *         return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeSimple other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_11dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_11dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 39, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":40
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeSimple other):
 *         return self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":39
 *         return self.departure_time >= other.departure_time and self.arrival_time_target <= other.arrival_time_target
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeSimple other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeSimple.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_11dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_11dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple, 1, "other", 0))) __PYX_ERR(0, 39, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_10dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_10dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":43
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_walk_duration;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 43, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 43, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L3_error)
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 43, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 43, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_12direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_12direct_walk_label(double __pyx_v_departure_time, double __pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":44
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_departure_time + __pyx_v_walk_duration)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":43
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":46
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_15duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_15duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 46, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":47
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTimeSimple get_copy(self):
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":46
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeSimple.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_15duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_15duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":49
 *         return self.arrival_time_target - self.departure_time
 * 
 *     cpdef LabelTimeSimple get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(self.departure_time, self.arrival_time_target)
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_17get_copy)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple))))) __PYX_ERR(0, 49, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":50
 * 
 *     cpdef LabelTimeSimple get_copy(self):
 *         return LabelTimeSimple(self.departure_time, self.arrival_time_target)             # <<<<<<<<<<<<<<
 * 
 * cdef class LabelTime:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":49
 *         return self.arrival_time_target - self.departure_time
 * 
 *     cpdef LabelTimeSimple get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(self.departure_time, self.arrival_time_target)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_16get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_get_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":6
 *     """
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public double arrival_time_target
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 6, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":7
 *     cdef:
 *         public double departure_time
 *         public double arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeSimple.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":61
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,             # <<<<<<<<<<<<<<
 *                  **kwargs):
 *         self.departure_time = departure_time
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_arrival_time_target;
  int __pyx_v_first_leg_is_walk;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_arrival_time_target,&__pyx_n_s_first_leg_is_walk,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrival_time_target);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_first_leg_is_walk);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    } else {
      __pyx_v_departure_time = __pyx_k__3;
    }
    if (values[1]) {
      __pyx_v_arrival_time_target = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_arrival_time_target == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    } else {
      __pyx_v_arrival_time_target = __pyx_k__4;
    }
    if (values[2]) {
      __pyx_v_first_leg_is_walk = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_first_leg_is_walk == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
    } else {
      __pyx_v_first_leg_is_walk = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), __pyx_v_departure_time, __pyx_v_arrival_time_target, __pyx_v_first_leg_is_walk, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_first_leg_is_walk, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":63
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,
 *                  **kwargs):
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.arrival_time_target = arrival_time_target
 *         self.first_leg_is_walk = first_leg_is_walk
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":64
 *                  **kwargs):
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target             # <<<<<<<<<<<<<<
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 */
  __pyx_v_self->arrival_time_target = __pyx_v_arrival_time_target;

  /* "gtfspy/routing/label.pyx":65
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 *         self.first_leg_is_walk = first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     cpdef _tuple_for_ordering(self):
 */
  __pyx_v_self->first_leg_is_walk = __pyx_v_first_leg_is_walk;

  /* "gtfspy/routing/label.pyx":61
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,             # <<<<<<<<<<<<<<
 *                  **kwargs):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":67
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     cpdef _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_3_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_9LabelTime__tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_3_tuple_for_ordering)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":68
 * 
 *     cpdef _tuple_for_ordering(self):
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelTime self, LabelTime other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(__pyx_v_self->first_leg_is_walk != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":67
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     cpdef _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_3_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_3_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_2_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_2_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_9LabelTime__tuple_for_ordering(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":70
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelTime self, LabelTime other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_5__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelTime, 1, "self", 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTime, 1, "other", 0))) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_4__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_4__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":71
 * 
 *     def __richcmp__(LabelTime self, LabelTime other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self->__pyx_vtab)->_tuple_for_ordering(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":72
 *     def __richcmp__(LabelTime self, LabelTime other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other->__pyx_vtab)->_tuple_for_ordering(__pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":73
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_2 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_2) {

    /* "gtfspy/routing/label.pyx":74
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":73
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":75
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_2 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_2) {

    /* "gtfspy/routing/label.pyx":76
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":75
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":77
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":78
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":77
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":79
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":80
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":79
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":81
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":82
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":81
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":83
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":84
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     # __getstate__ and __setstate__ : for pickling
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":83
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":70
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelTime self, LabelTime other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":87
 * 
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_7__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_6__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_6__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":88
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target, self.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":87
 * 
 *     # __getstate__ and __setstate__ : for pickling
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":90
 *         return self.departure_time, self.arrival_time_target, self.first_leg_is_walk
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_9__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_8__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_8__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  double __pyx_t_6;
  double __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":91
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(LabelTime self, LabelTime other) except *:
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 91, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 91, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->departure_time = __pyx_t_6;
  __pyx_v_self->arrival_time_target = __pyx_t_7;
  __pyx_v_self->first_leg_is_walk = __pyx_t_8;

  /* "gtfspy/routing/label.pyx":90
 *         return self.departure_time, self.arrival_time_target, self.first_leg_is_walk
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":93
 *         self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state
 * 
 *     cpdef int dominates(LabelTime self, LabelTime other) except *:             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 93, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":94
 * 
 *     cpdef int dominates(LabelTime self, LabelTime other) except *:
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self->__pyx_vtab)->_tuple_for_ordering(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":95
 *     cpdef int dominates(LabelTime self, LabelTime other) except *:
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other->__pyx_vtab)->_tuple_for_ordering(__pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":96
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(LabelTime self, LabelTime other):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self_tuple);
  __Pyx_GIVEREF(__pyx_v_self_tuple);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self_tuple);
  __Pyx_INCREF(__pyx_v_other_tuple);
  __Pyx_GIVEREF(__pyx_v_other_tuple);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_other_tuple);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 96, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 96, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 96, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 96, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 96, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 96, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 96, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":93
 *         self.departure_time, self.arrival_time_target, self.first_leg_is_walk = state
 * 
 *     cpdef int dominates(LabelTime self, LabelTime other) except *:             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTime, 1, "other", 0))) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates(__pyx_v_self, __pyx_v_other, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 93, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":98
 *         return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])
 * 
 *     cpdef int dominates_ignoring_dep_time(LabelTime self, LabelTime other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_13dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 98, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":99
 * 
 *     cpdef int dominates_ignoring_dep_time(LabelTime self, LabelTime other):
 *         return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTime other):
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":98
 *         return all([(s >= o) for s, o in zip(self_tuple, other_tuple)])
 * 
 *     cpdef int dominates_ignoring_dep_time(LabelTime self, LabelTime other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTime.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTime, 1, "other", 0))) __PYX_ERR(0, 98, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_12dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_12dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":101
 *         return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTime other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_15dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":102
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTime other):
 *         return self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTime get_copy(self):
 */
  __pyx_r = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":101
 *         return self.arrival_time_target <= other.arrival_time_target and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTime other):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTime.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTime, 1, "other", 0))) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_14dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_14dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":104
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 *     cpdef LabelTime get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17get_copy)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTime))))) __PYX_ERR(0, 104, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":105
 * 
 *     cpdef LabelTime get_copy(self):
 *         return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTime get_copy_with_specified_departure_time(self, double departure_time):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime), __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":104
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 *     cpdef LabelTime get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_16get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":107
 *         return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 *     cpdef LabelTime get_copy_with_specified_departure_time(self, double departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_departure_time); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_specified_departur); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19get_copy_with_specified_departure_time)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_departure_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 107, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTime))))) __PYX_ERR(0, 107, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":108
 * 
 *     cpdef LabelTime get_copy_with_specified_departure_time(self, double departure_time):
 *         return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime), __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":107
 *         return LabelTime(self.departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 *     cpdef LabelTime get_copy_with_specified_departure_time(self, double departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_arg_departure_time) {
  double __pyx_v_departure_time;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  assert(__pyx_arg_departure_time); {
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(__pyx_arg_departure_time); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 107, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_18get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((double)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_18get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, double __pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_specified_departure_time(__pyx_v_self, __pyx_v_departure_time, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":111
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_walk_duration;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 111, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 111, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L3_error)
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 111, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 111, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_20direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_20direct_walk_label(double __pyx_v_departure_time, double __pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":112
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTime(departure_time, departure_time + walk_duration, True)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((__pyx_v_departure_time + __pyx_v_walk_duration)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_3, 2, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":111
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":114
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_23duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_9LabelTime_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_23duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 114, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":115
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTime get_copy_with_walk_added(self, walk_duration):
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":114
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTime.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_23duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_23duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_22duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_22duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_9LabelTime_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":117
 *         return self.arrival_time_target - self.departure_time
 * 
 *     cpdef LabelTime get_copy_with_walk_added(self, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_25get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_v_walk_duration); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_walk_duration, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_walk_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_25get_copy_with_walk_added)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_walk_duration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_walk_duration};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_walk_duration};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_walk_duration);
          __Pyx_GIVEREF(__pyx_v_walk_duration);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_walk_duration);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTime))))) __PYX_ERR(0, 117, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":118
 * 
 *     cpdef LabelTime get_copy_with_walk_added(self, walk_duration):
 *         return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 * cdef class LabelTimeWithBoardingsCount:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_walk_duration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime), __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":117
 *         return self.arrival_time_target - self.departure_time
 * 
 *     cpdef LabelTime get_copy_with_walk_added(self, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(self.departure_time - walk_duration, self.arrival_time_target, self.first_leg_is_walk)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_25get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_v_walk_duration); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_25get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_24get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((PyObject *)__pyx_v_walk_duration));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_24get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_walk_added(__pyx_v_self, __pyx_v_walk_duration, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":57
 *     """
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public double arrival_time_target
 *         public bint first_leg_is_walk
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":58
 *     cdef:
 *         public double departure_time
 *         public double arrival_time_target             # <<<<<<<<<<<<<<
 *         public bint first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":59
 *         public double departure_time
 *         public double arrival_time_target
 *         public bint first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.first_leg_is_walk.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTime.first_leg_is_walk.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":128
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int n_boardings, bint first_leg_is_walk):
 *         self.departure_time = departure_time
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_arrival_time_target;
  int __pyx_v_n_boardings;
  int __pyx_v_first_leg_is_walk;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_arrival_time_target,&__pyx_n_s_n_boardings,&__pyx_n_s_first_leg_is_walk,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrival_time_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n_boardings)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 128, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_first_leg_is_walk)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 128, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 128, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    __pyx_v_arrival_time_target = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_arrival_time_target == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 128, __pyx_L3_error)
    __pyx_v_n_boardings = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n_boardings == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
    __pyx_v_first_leg_is_walk = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_first_leg_is_walk == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 128, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), __pyx_v_departure_time, __pyx_v_arrival_time_target, __pyx_v_n_boardings, __pyx_v_first_leg_is_walk);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_n_boardings, int __pyx_v_first_leg_is_walk) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":130
 *     def __init__(self, double departure_time, double arrival_time_target,
 *                  int n_boardings, bint first_leg_is_walk):
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":131
 *                  int n_boardings, bint first_leg_is_walk):
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target             # <<<<<<<<<<<<<<
 *         self.n_boardings = n_boardings
 *         self.first_leg_is_walk = first_leg_is_walk
 */
  __pyx_v_self->arrival_time_target = __pyx_v_arrival_time_target;

  /* "gtfspy/routing/label.pyx":132
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings             # <<<<<<<<<<<<<<
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 */
  __pyx_v_self->n_boardings = __pyx_v_n_boardings;

  /* "gtfspy/routing/label.pyx":133
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings
 *         self.first_leg_is_walk = first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_v_self->first_leg_is_walk = __pyx_v_first_leg_is_walk;

  /* "gtfspy/routing/label.pyx":128
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int n_boardings, bint first_leg_is_walk):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":135
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.n_boardings
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_2__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":136
 * 
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target, self.n_boardings             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":135
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.n_boardings
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":138
 *         return self.departure_time, self.arrival_time_target, self.n_boardings
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.n_boardings = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_4__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  double __pyx_t_6;
  double __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":139
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target, self.n_boardings = state             # <<<<<<<<<<<<<<
 * 
 *     def _tuple_for_ordering(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 139, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 139, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->departure_time = __pyx_t_6;
  __pyx_v_self->arrival_time_target = __pyx_t_7;
  __pyx_v_self->n_boardings = __pyx_t_8;

  /* "gtfspy/routing/label.pyx":138
 *         return self.departure_time, self.arrival_time_target, self.n_boardings
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.n_boardings = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":141
 *         self.departure_time, self.arrival_time_target, self.n_boardings = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_6_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);

  /* "gtfspy/routing/label.pyx":142
 * 
 *     def _tuple_for_ordering(self):
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int((-__pyx_v_self->n_boardings)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong((!(__pyx_v_self->first_leg_is_walk != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":141
 *         self.departure_time, self.arrival_time_target, self.n_boardings = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":144
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "self", 0))) __PYX_ERR(0, 144, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_8__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":145
 * 
 *     def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":146
 *     def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":147
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":148
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":147
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":149
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":150
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":149
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":151
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":152
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":151
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":153
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":154
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":153
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":155
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":156
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":155
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":157
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":158
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelTimeWithBoardingsCount other):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 158, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":157
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":144
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelTimeWithBoardingsCount self, LabelTimeWithBoardingsCount other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":160
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 160, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":173
 *             True if this ParetoTuple dominates the other, otherwise False
 *         """
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":174
 *         """
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":175
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeWithBoardingsCount other):
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self_tuple);
  __Pyx_GIVEREF(__pyx_v_self_tuple);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self_tuple);
  __Pyx_INCREF(__pyx_v_other_tuple);
  __Pyx_GIVEREF(__pyx_v_other_tuple);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_other_tuple);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 175, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_2);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 175, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 175, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 175, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 175, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 175, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = (!__pyx_t_11);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":160
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_10dominates[] = "\n        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles\n\n        Parameters\n        ----------\n        other: LabelTimeWithBoardingsCount\n\n        Returns\n        -------\n        dominates: bint\n            True if this ParetoTuple dominates the other, otherwise False\n        ";
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":177
 *         return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_dominates = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_13dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":179
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeWithBoardingsCount other):
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.n_boardings <= other.n_boardings
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":180
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 *             self.n_boardings <= other.n_boardings             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->n_boardings <= __pyx_v_other->n_boardings);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":182
 *             self.n_boardings <= other.n_boardings
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeWithBoardingsCount other):
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_dominates); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 182, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":177
 *         return not any([(s < o) for s, o in zip(self_tuple, other_tuple)])
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dominates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 177, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_12dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":184
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_15dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":188
 *             int dominates
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.n_boardings <= other.n_boardings and
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":189
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 *             self.n_boardings <= other.n_boardings and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->n_boardings <= __pyx_v_other->n_boardings);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":190
 *             self.arrival_time_target <= other.arrival_time_target and
 *             self.n_boardings <= other.n_boardings and
 *             self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":192
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeWithBoardingsCount other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":184
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":194
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17dominates_ignoring_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":198
 *             int dominates
 *         dominates = (
 *             self.n_boardings <= other.n_boardings and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->n_boardings <= __pyx_v_other->n_boardings);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":199
 *         dominates = (
 *             self.n_boardings <= other.n_boardings and
 *             self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":201
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeWithBoardingsCount other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":194
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_16dominates_ignoring_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":203
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_and); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19dominates_ignoring_dep_time_and_n_boardings)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 203, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":207
 *             int dominates
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":208
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 *             self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":210
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy(self):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":203
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeWithBoardingsCount other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, 1, "other", 0))) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_18dominates_ignoring_dep_time_and_n_boardings(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_and_n_boardings(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":212
 *         return dominates
 * 
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_21get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_21get_copy)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":213
 * 
 *     cpdef get_copy(self):
 *         return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":214
 *     cpdef get_copy(self):
 *         return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":213
 * 
 *     cpdef get_copy(self):
 *         return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":212
 *         return dominates
 * 
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_21get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_21get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_20get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_20get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":216
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_23get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_specified_departur); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_23get_copy_with_specified_departure_time)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_departure_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 216, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_departure_time);
          __Pyx_GIVEREF(__pyx_v_departure_time);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_departure_time);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":217
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":218
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":217
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount), __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":216
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.first_leg_is_walk)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_23get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_23get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_22get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_22get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_specified_departure_time(__pyx_v_self, __pyx_v_departure_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":220
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_25duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_25duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 220, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":221
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":220
 *                                            self.n_boardings, self.first_leg_is_walk)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeWithBoardingsCount.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_25duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_25duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_24duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_24duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":224
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_departure_time = 0;
  PyObject *__pyx_v_walk_duration = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 224, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 224, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = values[0];
    __pyx_v_walk_duration = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 224, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_26direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_26direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":225
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_departure_time, __pyx_v_walk_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_0);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":224
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":227
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, True)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_29get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_arg_walk_duration); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_walk_duration, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_walk_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_29get_copy_with_walk_added)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_walk_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 227, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount))))) __PYX_ERR(0, 227, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":228
 * 
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):
 *         return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.n_boardings, True)
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->departure_time - __pyx_v_walk_duration)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":229
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):
 *         return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, True)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":228
 * 
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):
 *         return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.n_boardings, True)
 * 
 */
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_6, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount), __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":227
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeWithBoardingsCount get_copy_with_walk_added(self, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_29get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_arg_walk_duration); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_29get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_arg_walk_duration) {
  double __pyx_v_walk_duration;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  assert(__pyx_arg_walk_duration); {
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(__pyx_arg_walk_duration); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_28get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((double)__pyx_v_walk_duration));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_28get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, double __pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_walk_added(__pyx_v_self, __pyx_v_walk_duration, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":231
 *                                            self.arrival_time_target, self.n_boardings, True)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.first_leg_is_walk))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_31__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_31__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_30__str__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_30__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gtfspy/routing/label.pyx":232
 * 
 *     def __str__(self):
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.first_leg_is_walk))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":231
 *                                            self.arrival_time_target, self.n_boardings, True)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.first_leg_is_walk))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":122
 * cdef class LabelTimeWithBoardingsCount:
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public double arrival_time_target
 *         public int n_boardings
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":123
 *     cdef:
 *         public double departure_time
 *         public double arrival_time_target             # <<<<<<<<<<<<<<
 *         public int n_boardings
 *         public bint first_leg_is_walk
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 123, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":124
 *         public double departure_time
 *         public double arrival_time_target
 *         public int n_boardings             # <<<<<<<<<<<<<<
 *         public bint first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.n_boardings.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_v_self->n_boardings = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.n_boardings.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":125
 *         public double arrival_time_target
 *         public int n_boardings
 *         public bint first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.first_leg_is_walk.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 125, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeWithBoardingsCount.first_leg_is_walk.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":241
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):             # <<<<<<<<<<<<<<
 *         self.n_boardings = n_boardings
 *         self.departure_time = departure_time
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_n_boardings = 0;
  double __pyx_v_departure_time;
  PyObject *__pyx_v_first_leg_is_walk = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return -1;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_n_boardings,&__pyx_n_s_departure_time,&__pyx_n_s_first_leg_is_walk,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject *)__pyx_int_0);
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n_boardings);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_first_leg_is_walk);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 241, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_n_boardings = values[0];
    if (values[1]) {
      __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 241, __pyx_L3_error)
    } else {
      __pyx_v_departure_time = __pyx_k__5;
    }
    __pyx_v_first_leg_is_walk = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 241, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), __pyx_v_n_boardings, __pyx_v_departure_time, __pyx_v_first_leg_is_walk, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_n_boardings, double __pyx_v_departure_time, PyObject *__pyx_v_first_leg_is_walk, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":242
 * 
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):
 *         self.n_boardings = n_boardings             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.first_leg_is_walk = first_leg_is_walk
 */
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_n_boardings); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 242, __pyx_L1_error)
  __pyx_v_self->n_boardings = __pyx_t_1;

  /* "gtfspy/routing/label.pyx":243
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):
 *         self.n_boardings = n_boardings
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":244
 *         self.n_boardings = n_boardings
 *         self.departure_time = departure_time
 *         self.first_leg_is_walk = first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_first_leg_is_walk); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 244, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_2;

  /* "gtfspy/routing/label.pyx":241
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):             # <<<<<<<<<<<<<<
 *         self.n_boardings = n_boardings
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":246
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.n_boardings, self.first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_2__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":247
 * 
 *     def __getstate__(self):
 *         return self.departure_time, self.n_boardings, self.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":246
 *         self.first_leg_is_walk = first_leg_is_walk
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.n_boardings, self.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":249
 *         return self.departure_time, self.n_boardings, self.first_leg_is_walk
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.n_boardings, self.first_leg_is_walk = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_4__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  double __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":250
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.n_boardings, self.first_leg_is_walk = state             # <<<<<<<<<<<<<<
 * 
 *     def _tuple_for_ordering(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 250, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 2; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 3) < 0) __PYX_ERR(0, 250, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 250, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->departure_time = __pyx_t_6;
  __pyx_v_self->n_boardings = __pyx_t_7;
  __pyx_v_self->first_leg_is_walk = __pyx_t_8;

  /* "gtfspy/routing/label.pyx":249
 *         return self.departure_time, self.n_boardings, self.first_leg_is_walk
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.n_boardings, self.first_leg_is_walk = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":252
 *         self.departure_time, self.n_boardings, self.first_leg_is_walk = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return -self.n_boardings, self.departure_time, not self.first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_6_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);

  /* "gtfspy/routing/label.pyx":253
 * 
 *     def _tuple_for_ordering(self):
 *         return -self.n_boardings, self.departure_time, not self.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int((-__pyx_v_self->n_boardings)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(__pyx_v_self->first_leg_is_walk != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":252
 *         self.departure_time, self.n_boardings, self.first_leg_is_walk = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return -self.n_boardings, self.departure_time, not self.first_leg_is_walk
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":255
 *         return -self.n_boardings, self.departure_time, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, 1, "self", 0))) __PYX_ERR(0, 255, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, 1, "other", 0))) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_8__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":256
 * 
 *     def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 256, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":257
 *     def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:
 *             return self_tuple == other_tuple
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":258
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":259
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":258
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:
 */
  }

  /* "gtfspy/routing/label.pyx":260
 *         if op == 2:
 *             return self_tuple == other_tuple
 *         if op == 3:             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":261
 *             return self_tuple == other_tuple
 *         if op == 3:
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":260
 *         if op == 2:
 *             return self_tuple == other_tuple
 *         if op == 3:             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":262
 *         if op == 3:
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":263
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":262
 *         if op == 3:
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":264
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":265
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 265, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":264
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":266
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":267
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":266
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":268
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":269
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelVehLegCount other) except *:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":268
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":255
 *         return -self.n_boardings, self.departure_time, not self.first_leg_is_walk
 * 
 *     def __richcmp__(LabelVehLegCount self, LabelVehLegCount other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":271
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelVehLegCount other) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 271, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":284
 *             True if this ParetoTuple dominates the other, otherwise False
 *         """
 *         return self.n_boardings <= other.n_boardings and self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelVehLegCount other):
 */
  __pyx_t_7 = (__pyx_v_self->n_boardings <= __pyx_v_other->n_boardings);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":271
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelVehLegCount other) except *:             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_16LabelVehLegCount_10dominates[] = "\n        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles\n\n        Parameters\n        ----------\n        other: LabelTimeWithBoardingsCount\n\n        Returns\n        -------\n        dominates: bint\n            True if this ParetoTuple dominates the other, otherwise False\n        ";
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, 1, "other", 0))) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates(__pyx_v_self, __pyx_v_other, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":286
 *         return self.n_boardings <= other.n_boardings and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelVehLegCount other):             # <<<<<<<<<<<<<<
 *         return self.dominates(other)
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_13dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":287
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelVehLegCount other):
 *         return self.dominates(other)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelVehLegCount other):
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self->__pyx_vtab)->dominates(__pyx_v_self, __pyx_v_other, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 287, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":286
 *         return self.n_boardings <= other.n_boardings and self.first_leg_is_walk <= other.first_leg_is_walk
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelVehLegCount other):             # <<<<<<<<<<<<<<
 *         return self.dominates(other)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelVehLegCount.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_13dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, 1, "other", 0))) __PYX_ERR(0, 286, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_12dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_12dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":289
 *         return self.dominates(other)
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelVehLegCount other):             # <<<<<<<<<<<<<<
 *         return self.n_boardings <= other.n_boardings
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_15dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":290
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelVehLegCount other):
 *         return self.n_boardings <= other.n_boardings             # <<<<<<<<<<<<<<
 * 
 *     def get_copy(self):
 */
  __pyx_r = (__pyx_v_self->n_boardings <= __pyx_v_other->n_boardings);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":289
 *         return self.dominates(other)
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelVehLegCount other):             # <<<<<<<<<<<<<<
 *         return self.n_boardings <= other.n_boardings
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelVehLegCount.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_15dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, 1, "other", 0))) __PYX_ERR(0, 289, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":292
 *         return self.n_boardings <= other.n_boardings
 * 
 *     def get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_16get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_16get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);

  /* "gtfspy/routing/label.pyx":293
 * 
 *     def get_copy(self):
 *         return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     def get_copy_with_specified_departure_time(self, departure_time):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_first_leg_is_walk, __pyx_t_3) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount), __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":292
 *         return self.n_boardings <= other.n_boardings
 * 
 *     def get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":295
 *         return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)
 * 
 *     def get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_19get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_19get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_18get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_18get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);

  /* "gtfspy/routing/label.pyx":296
 * 
 *     def get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)             # <<<<<<<<<<<<<<
 * 
 *     def get_copy_with_walk_added(self, walk_duration):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":295
 *         return LabelVehLegCount(self.n_boardings, first_leg_is_walk=self.first_leg_is_walk)
 * 
 *     def get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":298
 *         return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)
 * 
 *     def get_copy_with_walk_added(self, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,
 *                                 first_leg_is_walk=True)
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_21get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_v_walk_duration); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_21get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_20get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_walk_duration));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_20get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);

  /* "gtfspy/routing/label.pyx":299
 * 
 *     def get_copy_with_walk_added(self, walk_duration):
 *         return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                 first_leg_is_walk=True)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Subtract(__pyx_t_3, __pyx_v_walk_duration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_departure_time, __pyx_t_4) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":300
 *     def get_copy_with_walk_added(self, walk_duration):
 *         return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,
 *                                 first_leg_is_walk=True)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_first_leg_is_walk, Py_True) < 0) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":299
 * 
 *     def get_copy_with_walk_added(self, walk_duration):
 *         return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                 first_leg_is_walk=True)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount), __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":298
 *         return LabelVehLegCount(self.n_boardings, departure_time, self.first_leg_is_walk)
 * 
 *     def get_copy_with_walk_added(self, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(self.n_boardings, departure_time=self.departure_time - walk_duration,
 *                                 first_leg_is_walk=True)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":303
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(0, departure_time, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_departure_time = 0;
  CYTHON_UNUSED PyObject *__pyx_v_walk_duration = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 303, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 303, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = values[0];
    __pyx_v_walk_duration = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 303, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_22direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_22direct_walk_label(PyObject *__pyx_v_departure_time, CYTHON_UNUSED PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":304
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelVehLegCount(0, departure_time, True)             # <<<<<<<<<<<<<<
 * 
 * # ctypedef fused label:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_departure_time);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_1, 2, Py_True);
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":303
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(0, departure_time, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":237
 * cdef class LabelVehLegCount:
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public int n_boardings
 *         public bint first_leg_is_walk
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":238
 *     cdef:
 *         public double departure_time
 *         public int n_boardings             # <<<<<<<<<<<<<<
 *         public bint first_leg_is_walk
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.n_boardings.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 238, __pyx_L1_error)
  __pyx_v_self->n_boardings = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.n_boardings.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":239
 *         public double departure_time
 *         public int n_boardings
 *         public bint first_leg_is_walk             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.first_leg_is_walk.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelVehLegCount.first_leg_is_walk.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":314
 * #     return first.dominates(other)
 * 
 * def compute_pareto_front_smart(list label_list):             # <<<<<<<<<<<<<<
 *     return compute_pareto_front(label_list)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_1compute_pareto_front_smart(PyObject *__pyx_self, PyObject *__pyx_v_label_list); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_1compute_pareto_front_smart = {"compute_pareto_front_smart", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_1compute_pareto_front_smart, METH_O, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_1compute_pareto_front_smart(PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_pareto_front_smart (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_label_list), (&PyList_Type), 1, "label_list", 1))) __PYX_ERR(0, 314, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_compute_pareto_front_smart(__pyx_self, ((PyObject*)__pyx_v_label_list));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_compute_pareto_front_smart(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("compute_pareto_front_smart", 0);

  /* "gtfspy/routing/label.pyx":315
 * 
 * def compute_pareto_front_smart(list label_list):
 *     return compute_pareto_front(label_list)             # <<<<<<<<<<<<<<
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_compute_pareto_front); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_label_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_label_list};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_v_label_list};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_v_label_list);
      __Pyx_GIVEREF(__pyx_v_label_list);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_label_list);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":314
 * #     return first.dominates(other)
 * 
 * def compute_pareto_front_smart(list label_list):             # <<<<<<<<<<<<<<
 *     return compute_pareto_front(label_list)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.compute_pareto_front_smart", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":317
 *     return compute_pareto_front(label_list)
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):             # <<<<<<<<<<<<<<
 *     pareto_front = []
 *     if len(label_list) == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_3compute_pareto_front(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_3compute_pareto_front = {"compute_pareto_front", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_3compute_pareto_front, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_3compute_pareto_front(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_label_list = 0;
  PyObject *__pyx_v_finalization = 0;
  PyObject *__pyx_v_ignore_n_boardings = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_pareto_front (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_label_list,&__pyx_n_s_finalization,&__pyx_n_s_ignore_n_boardings,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)Py_False);
    values[2] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_label_list)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_finalization);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ignore_n_boardings);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "compute_pareto_front") < 0)) __PYX_ERR(0, 317, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_label_list = ((PyObject*)values[0]);
    __pyx_v_finalization = values[1];
    __pyx_v_ignore_n_boardings = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("compute_pareto_front", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 317, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.compute_pareto_front", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_label_list), (&PyList_Type), 1, "label_list", 1))) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_2compute_pareto_front(__pyx_self, __pyx_v_label_list, __pyx_v_finalization, __pyx_v_ignore_n_boardings);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_2compute_pareto_front(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list, PyObject *__pyx_v_finalization, PyObject *__pyx_v_ignore_n_boardings) {
  PyObject *__pyx_v_pareto_front = NULL;
  PyObject *__pyx_v_label = NULL;
  PyObject *__pyx_v_dominates = NULL;
  PyObject *__pyx_v_current_best_labels_wo_deptime = NULL;
  PyObject *__pyx_v_new_label = NULL;
  int __pyx_v_is_dominated;
  PyObject *__pyx_v_best_label = NULL;
  PyObject *__pyx_v_new_best = NULL;
  PyObject *__pyx_v_old_partial_best = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("compute_pareto_front", 0);
  __Pyx_INCREF(__pyx_v_label_list);

  /* "gtfspy/routing/label.pyx":318
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):
 *     pareto_front = []             # <<<<<<<<<<<<<<
 *     if len(label_list) == 0:
 *         return pareto_front
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pareto_front = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":319
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):
 *     pareto_front = []
 *     if len(label_list) == 0:             # <<<<<<<<<<<<<<
 *         return pareto_front
 * 
 */
  if (unlikely(__pyx_v_label_list == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 319, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_label_list); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 == 0) != 0);
  if (__pyx_t_3) {

    /* "gtfspy/routing/label.pyx":320
 *     pareto_front = []
 *     if len(label_list) == 0:
 *         return pareto_front             # <<<<<<<<<<<<<<
 * 
 *     assert(not (finalization and ignore_n_boardings))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_pareto_front);
    __pyx_r = __pyx_v_pareto_front;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":319
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):
 *     pareto_front = []
 *     if len(label_list) == 0:             # <<<<<<<<<<<<<<
 *         return pareto_front
 * 
 */
  }

  /* "gtfspy/routing/label.pyx":322
 *         return pareto_front
 * 
 *     assert(not (finalization and ignore_n_boardings))             # <<<<<<<<<<<<<<
 *     # determine function used for domination:
 *     label = next(iter(label_list))
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_finalization); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_ignore_n_boardings); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 322, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    if (unlikely(!((!__pyx_t_3) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 322, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":324
 *     assert(not (finalization and ignore_n_boardings))
 *     # determine function used for domination:
 *     label = next(iter(label_list))             # <<<<<<<<<<<<<<
 *     if finalization:
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_label_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyIter_Next(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_label = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":325
 *     # determine function used for domination:
 *     label = next(iter(label_list))
 *     if finalization:             # <<<<<<<<<<<<<<
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 *     elif ignore_n_boardings:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_finalization); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 325, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "gtfspy/routing/label.pyx":326
 *     label = next(iter(label_list))
 *     if finalization:
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization             # <<<<<<<<<<<<<<
 *     elif ignore_n_boardings:
 *         dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_class); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dominates = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "gtfspy/routing/label.pyx":325
 *     # determine function used for domination:
 *     label = next(iter(label_list))
 *     if finalization:             # <<<<<<<<<<<<<<
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 *     elif ignore_n_boardings:
 */
    goto __pyx_L6;
  }

  /* "gtfspy/routing/label.pyx":327
 *     if finalization:
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 *     elif ignore_n_boardings:             # <<<<<<<<<<<<<<
 *         dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings
 *     else:
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_ignore_n_boardings); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 327, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "gtfspy/routing/label.pyx":328
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 *     elif ignore_n_boardings:
 *         dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings             # <<<<<<<<<<<<<<
 *     else:
 *         dominates = label.__class__.dominates_ignoring_dep_time
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dominates_ignoring_dep_time_and); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_dominates = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "gtfspy/routing/label.pyx":327
 *     if finalization:
 *         dominates = label.__class__.dominates_ignoring_dep_time_finalization
 *     elif ignore_n_boardings:             # <<<<<<<<<<<<<<
 *         dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings
 *     else:
 */
    goto __pyx_L6;
  }

  /* "gtfspy/routing/label.pyx":330
 *         dominates = label.__class__.dominates_ignoring_dep_time_and_n_boardings
 *     else:
 *         dominates = label.__class__.dominates_ignoring_dep_time             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_class); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dominates = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L6:;

  /* "gtfspy/routing/label.pyx":333
 * 
 * 
 *     label_list = list(reversed(sorted(label_list)))  # n log(n)             # <<<<<<<<<<<<<<
 *     # assume only that label_list is sorted by departure time (best last)
 *     current_best_labels_wo_deptime = []
 */
  __pyx_t_5 = PySequence_List(__pyx_v_label_list); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_6 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_reversed, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_label_list, ((PyObject*)__pyx_t_5));
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":335
 *     label_list = list(reversed(sorted(label_list)))  # n log(n)
 *     # assume only that label_list is sorted by departure time (best last)
 *     current_best_labels_wo_deptime = []             # <<<<<<<<<<<<<<
 *     for new_label in label_list:  # n times
 *         is_dominated = False
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_current_best_labels_wo_deptime = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":336
 *     # assume only that label_list is sorted by departure time (best last)
 *     current_best_labels_wo_deptime = []
 *     for new_label in label_list:  # n times             # <<<<<<<<<<<<<<
 *         is_dominated = False
 *         for best_label in current_best_labels_wo_deptime:
 */
  __pyx_t_5 = __pyx_v_label_list; __Pyx_INCREF(__pyx_t_5); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_5)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_2); __Pyx_INCREF(__pyx_t_1); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 336, __pyx_L1_error)
    #else
    __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_new_label, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gtfspy/routing/label.pyx":337
 *     current_best_labels_wo_deptime = []
 *     for new_label in label_list:  # n times
 *         is_dominated = False             # <<<<<<<<<<<<<<
 *         for best_label in current_best_labels_wo_deptime:
 *             # the size of current_best_labels_wo_deptime should remain small
 */
    __pyx_v_is_dominated = 0;

    /* "gtfspy/routing/label.pyx":338
 *     for new_label in label_list:  # n times
 *         is_dominated = False
 *         for best_label in current_best_labels_wo_deptime:             # <<<<<<<<<<<<<<
 *             # the size of current_best_labels_wo_deptime should remain small
 *             # the new_label can dominate the old ones only partially
 */
    __pyx_t_1 = __pyx_v_current_best_labels_wo_deptime; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
      #else
      __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 338, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_best_label, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "gtfspy/routing/label.pyx":342
 *             # the new_label can dominate the old ones only partially
 *             # check if the new one is dominated by the old ones ->
 *             if dominates(best_label, new_label):             # <<<<<<<<<<<<<<
 *                 is_dominated = True
 *                 break
 */
      __Pyx_INCREF(__pyx_v_dominates);
      __pyx_t_9 = __pyx_v_dominates; __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
          __pyx_t_11 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_best_label, __pyx_v_new_label};
        __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
        PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_v_best_label, __pyx_v_new_label};
        __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else
      #endif
      {
        __pyx_t_12 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_v_best_label);
        __Pyx_GIVEREF(__pyx_v_best_label);
        PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_v_best_label);
        __Pyx_INCREF(__pyx_v_new_label);
        __Pyx_GIVEREF(__pyx_v_new_label);
        PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_new_label);
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 342, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_3) {

        /* "gtfspy/routing/label.pyx":343
 *             # check if the new one is dominated by the old ones ->
 *             if dominates(best_label, new_label):
 *                 is_dominated = True             # <<<<<<<<<<<<<<
 *                 break
 *         if is_dominated:
 */
        __pyx_v_is_dominated = 1;

        /* "gtfspy/routing/label.pyx":344
 *             if dominates(best_label, new_label):
 *                 is_dominated = True
 *                 break             # <<<<<<<<<<<<<<
 *         if is_dominated:
 *             continue  # do nothing
 */
        goto __pyx_L10_break;

        /* "gtfspy/routing/label.pyx":342
 *             # the new_label can dominate the old ones only partially
 *             # check if the new one is dominated by the old ones ->
 *             if dominates(best_label, new_label):             # <<<<<<<<<<<<<<
 *                 is_dominated = True
 *                 break
 */
      }

      /* "gtfspy/routing/label.pyx":338
 *     for new_label in label_list:  # n times
 *         is_dominated = False
 *         for best_label in current_best_labels_wo_deptime:             # <<<<<<<<<<<<<<
 *             # the size of current_best_labels_wo_deptime should remain small
 *             # the new_label can dominate the old ones only partially
 */
    }
    __pyx_L10_break:;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "gtfspy/routing/label.pyx":345
 *                 is_dominated = True
 *                 break
 *         if is_dominated:             # <<<<<<<<<<<<<<
 *             continue  # do nothing
 *         else:
 */
    __pyx_t_3 = (__pyx_v_is_dominated != 0);
    if (__pyx_t_3) {

      /* "gtfspy/routing/label.pyx":346
 *                 break
 *         if is_dominated:
 *             continue  # do nothing             # <<<<<<<<<<<<<<
 *         else:
 *             pareto_front.append(new_label)
 */
      goto __pyx_L7_continue;

      /* "gtfspy/routing/label.pyx":345
 *                 is_dominated = True
 *                 break
 *         if is_dominated:             # <<<<<<<<<<<<<<
 *             continue  # do nothing
 *         else:
 */
    }

    /* "gtfspy/routing/label.pyx":348
 *             continue  # do nothing
 *         else:
 *             pareto_front.append(new_label)             # <<<<<<<<<<<<<<
 *             new_best = []
 *             for old_partial_best in current_best_labels_wo_deptime:
 */
    /*else*/ {
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_pareto_front, __pyx_v_new_label); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 348, __pyx_L1_error)

      /* "gtfspy/routing/label.pyx":349
 *         else:
 *             pareto_front.append(new_label)
 *             new_best = []             # <<<<<<<<<<<<<<
 *             for old_partial_best in current_best_labels_wo_deptime:
 *                 if not dominates(new_label, old_partial_best):
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_new_best, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "gtfspy/routing/label.pyx":350
 *             pareto_front.append(new_label)
 *             new_best = []
 *             for old_partial_best in current_best_labels_wo_deptime:             # <<<<<<<<<<<<<<
 *                 if not dominates(new_label, old_partial_best):
 *                     new_best.append(old_partial_best)
 */
      __pyx_t_1 = __pyx_v_current_best_labels_wo_deptime; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      for (;;) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 350, __pyx_L1_error)
        #else
        __pyx_t_8 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_old_partial_best, __pyx_t_8);
        __pyx_t_8 = 0;

        /* "gtfspy/routing/label.pyx":351
 *             new_best = []
 *             for old_partial_best in current_best_labels_wo_deptime:
 *                 if not dominates(new_label, old_partial_best):             # <<<<<<<<<<<<<<
 *                     new_best.append(old_partial_best)
 *             new_best.append(new_label)
 */
        __Pyx_INCREF(__pyx_v_dominates);
        __pyx_t_9 = __pyx_v_dominates; __pyx_t_12 = NULL;
        __pyx_t_11 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_11 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_new_label, __pyx_v_old_partial_best};
          __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 351, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_8);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_9)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_new_label, __pyx_v_old_partial_best};
          __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_9, __pyx_temp+1-__pyx_t_11, 2+__pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 351, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_8);
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 351, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(__pyx_v_new_label);
          __Pyx_GIVEREF(__pyx_v_new_label);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_11, __pyx_v_new_label);
          __Pyx_INCREF(__pyx_v_old_partial_best);
          __Pyx_GIVEREF(__pyx_v_old_partial_best);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_11, __pyx_v_old_partial_best);
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 351, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 351, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_4 = ((!__pyx_t_3) != 0);
        if (__pyx_t_4) {

          /* "gtfspy/routing/label.pyx":352
 *             for old_partial_best in current_best_labels_wo_deptime:
 *                 if not dominates(new_label, old_partial_best):
 *                     new_best.append(old_partial_best)             # <<<<<<<<<<<<<<
 *             new_best.append(new_label)
 *             current_best_labels_wo_deptime = new_best
 */
          __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_new_best, __pyx_v_old_partial_best); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 352, __pyx_L1_error)

          /* "gtfspy/routing/label.pyx":351
 *             new_best = []
 *             for old_partial_best in current_best_labels_wo_deptime:
 *                 if not dominates(new_label, old_partial_best):             # <<<<<<<<<<<<<<
 *                     new_best.append(old_partial_best)
 *             new_best.append(new_label)
 */
        }

        /* "gtfspy/routing/label.pyx":350
 *             pareto_front.append(new_label)
 *             new_best = []
 *             for old_partial_best in current_best_labels_wo_deptime:             # <<<<<<<<<<<<<<
 *                 if not dominates(new_label, old_partial_best):
 *                     new_best.append(old_partial_best)
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "gtfspy/routing/label.pyx":353
 *                 if not dominates(new_label, old_partial_best):
 *                     new_best.append(old_partial_best)
 *             new_best.append(new_label)             # <<<<<<<<<<<<<<
 *             current_best_labels_wo_deptime = new_best
 *     return pareto_front
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_new_best, __pyx_v_new_label); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 353, __pyx_L1_error)

      /* "gtfspy/routing/label.pyx":354
 *                     new_best.append(old_partial_best)
 *             new_best.append(new_label)
 *             current_best_labels_wo_deptime = new_best             # <<<<<<<<<<<<<<
 *     return pareto_front
 * 
 */
      __Pyx_INCREF(__pyx_v_new_best);
      __Pyx_DECREF_SET(__pyx_v_current_best_labels_wo_deptime, __pyx_v_new_best);
    }

    /* "gtfspy/routing/label.pyx":336
 *     # assume only that label_list is sorted by departure time (best last)
 *     current_best_labels_wo_deptime = []
 *     for new_label in label_list:  # n times             # <<<<<<<<<<<<<<
 *         is_dominated = False
 *         for best_label in current_best_labels_wo_deptime:
 */
    __pyx_L7_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":355
 *             new_best.append(new_label)
 *             current_best_labels_wo_deptime = new_best
 *     return pareto_front             # <<<<<<<<<<<<<<
 * 
 * def compute_pareto_front_naive(list label_list):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pareto_front);
  __pyx_r = __pyx_v_pareto_front;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":317
 *     return compute_pareto_front(label_list)
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):             # <<<<<<<<<<<<<<
 *     pareto_front = []
 *     if len(label_list) == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("gtfspy.routing.label.compute_pareto_front", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pareto_front);
  __Pyx_XDECREF(__pyx_v_label);
  __Pyx_XDECREF(__pyx_v_dominates);
  __Pyx_XDECREF(__pyx_v_current_best_labels_wo_deptime);
  __Pyx_XDECREF(__pyx_v_new_label);
  __Pyx_XDECREF(__pyx_v_best_label);
  __Pyx_XDECREF(__pyx_v_new_best);
  __Pyx_XDECREF(__pyx_v_old_partial_best);
  __Pyx_XDECREF(__pyx_v_label_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":357
 *     return pareto_front
 * 
 * def compute_pareto_front_naive(list label_list):             # <<<<<<<<<<<<<<
 *     """
 *     Computes the Pareto frontier of a given label_list
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_5compute_pareto_front_naive(PyObject *__pyx_self, PyObject *__pyx_v_label_list); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_4compute_pareto_front_naive[] = "\n    Computes the Pareto frontier of a given label_list\n\n    Parameters\n    ----------\n    label_list: list[LabelTime]\n        (Or any list of objects, for which a function label.dominates(other) is defined.\n\n    Returns\n    -------\n    pareto_front: list[LabelTime]\n        List of labels that belong to the Pareto front.\n\n    Notes\n    -----\n    Code adapted from:\n    http://stackoverflow.com/questions/32791911/fast-calculation-of-pareto-front-in-python\n    ";
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_5compute_pareto_front_naive = {"compute_pareto_front_naive", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_5compute_pareto_front_naive, METH_O, __pyx_doc_6gtfspy_7routing_5label_4compute_pareto_front_naive};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_5compute_pareto_front_naive(PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("compute_pareto_front_naive (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_label_list), (&PyList_Type), 1, "label_list", 1))) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_4compute_pareto_front_naive(__pyx_self, ((PyObject*)__pyx_v_label_list));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_4compute_pareto_front_naive(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_v_dominated = NULL;
  PyObject *__pyx_v_pareto_front = NULL;
  PyObject *__pyx_v_remaining = NULL;
  PyObject *__pyx_v_candidate = NULL;
  PyObject *__pyx_v_new_remaining = NULL;
  int __pyx_v_is_dominated;
  Py_ssize_t __pyx_v_other_i;
  PyObject *__pyx_v_other = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("compute_pareto_front_naive", 0);

  /* "gtfspy/routing/label.pyx":382
 *     #    Label other
 *     #    Label candidate
 *     dominated = []             # <<<<<<<<<<<<<<
 *     pareto_front = []
 *     remaining = label_list
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dominated = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":383
 *     #    Label candidate
 *     dominated = []
 *     pareto_front = []             # <<<<<<<<<<<<<<
 *     remaining = label_list
 *     while remaining:  # (is not empty)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pareto_front = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":384
 *     dominated = []
 *     pareto_front = []
 *     remaining = label_list             # <<<<<<<<<<<<<<
 *     while remaining:  # (is not empty)
 *         candidate = remaining[0]
 */
  __Pyx_INCREF(__pyx_v_label_list);
  __pyx_v_remaining = __pyx_v_label_list;

  /* "gtfspy/routing/label.pyx":385
 *     pareto_front = []
 *     remaining = label_list
 *     while remaining:  # (is not empty)             # <<<<<<<<<<<<<<
 *         candidate = remaining[0]
 *         new_remaining = []
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_remaining != Py_None) && (PyList_GET_SIZE(__pyx_v_remaining) != 0);
    if (!__pyx_t_2) break;

    /* "gtfspy/routing/label.pyx":386
 *     remaining = label_list
 *     while remaining:  # (is not empty)
 *         candidate = remaining[0]             # <<<<<<<<<<<<<<
 *         new_remaining = []
 *         is_dominated = False
 */
    if (unlikely(__pyx_v_remaining == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 386, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_remaining, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_candidate, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "gtfspy/routing/label.pyx":387
 *     while remaining:  # (is not empty)
 *         candidate = remaining[0]
 *         new_remaining = []             # <<<<<<<<<<<<<<
 *         is_dominated = False
 *         for other_i in range(1, len(remaining)):
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_new_remaining, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "gtfspy/routing/label.pyx":388
 *         candidate = remaining[0]
 *         new_remaining = []
 *         is_dominated = False             # <<<<<<<<<<<<<<
 *         for other_i in range(1, len(remaining)):
 *             other = remaining[other_i]
 */
    __pyx_v_is_dominated = 0;

    /* "gtfspy/routing/label.pyx":389
 *         new_remaining = []
 *         is_dominated = False
 *         for other_i in range(1, len(remaining)):             # <<<<<<<<<<<<<<
 *             other = remaining[other_i]
 *             if candidate.dominates(other):
 */
    if (unlikely(__pyx_v_remaining == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 389, __pyx_L1_error)
    }
    __pyx_t_3 = PyList_GET_SIZE(__pyx_v_remaining); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 389, __pyx_L1_error)
    for (__pyx_t_4 = 1; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_other_i = __pyx_t_4;

      /* "gtfspy/routing/label.pyx":390
 *         is_dominated = False
 *         for other_i in range(1, len(remaining)):
 *             other = remaining[other_i]             # <<<<<<<<<<<<<<
 *             if candidate.dominates(other):
 *                 dominated.append(other)
 */
      if (unlikely(__pyx_v_remaining == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 390, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_remaining, __pyx_v_other_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_other, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "gtfspy/routing/label.pyx":391
 *         for other_i in range(1, len(remaining)):
 *             other = remaining[other_i]
 *             if candidate.dominates(other):             # <<<<<<<<<<<<<<
 *                 dominated.append(other)
 *             else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_candidate, __pyx_n_s_dominates); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_other};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_other};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 391, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_other);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "gtfspy/routing/label.pyx":392
 *             other = remaining[other_i]
 *             if candidate.dominates(other):
 *                 dominated.append(other)             # <<<<<<<<<<<<<<
 *             else:
 *                 new_remaining.append(other)
 */
        __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_dominated, __pyx_v_other); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 392, __pyx_L1_error)

        /* "gtfspy/routing/label.pyx":391
 *         for other_i in range(1, len(remaining)):
 *             other = remaining[other_i]
 *             if candidate.dominates(other):             # <<<<<<<<<<<<<<
 *                 dominated.append(other)
 *             else:
 */
        goto __pyx_L7;
      }

      /* "gtfspy/routing/label.pyx":394
 *                 dominated.append(other)
 *             else:
 *                 new_remaining.append(other)             # <<<<<<<<<<<<<<
 *                 if other.dominates(candidate):
 *                     is_dominated = True
 */
      /*else*/ {
        __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_new_remaining, __pyx_v_other); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 394, __pyx_L1_error)

        /* "gtfspy/routing/label.pyx":395
 *             else:
 *                 new_remaining.append(other)
 *                 if other.dominates(candidate):             # <<<<<<<<<<<<<<
 *                     is_dominated = True
 *         if is_dominated:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_dominates); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_candidate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_candidate};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
            PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_v_candidate};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
            __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 395, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(__pyx_v_candidate);
            __Pyx_GIVEREF(__pyx_v_candidate);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_candidate);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 395, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_2) {

          /* "gtfspy/routing/label.pyx":396
 *                 new_remaining.append(other)
 *                 if other.dominates(candidate):
 *                     is_dominated = True             # <<<<<<<<<<<<<<
 *         if is_dominated:
 *             dominated.append(candidate)
 */
          __pyx_v_is_dominated = 1;

          /* "gtfspy/routing/label.pyx":395
 *             else:
 *                 new_remaining.append(other)
 *                 if other.dominates(candidate):             # <<<<<<<<<<<<<<
 *                     is_dominated = True
 *         if is_dominated:
 */
        }
      }
      __pyx_L7:;
    }

    /* "gtfspy/routing/label.pyx":397
 *                 if other.dominates(candidate):
 *                     is_dominated = True
 *         if is_dominated:             # <<<<<<<<<<<<<<
 *             dominated.append(candidate)
 *         else:
 */
    __pyx_t_2 = (__pyx_v_is_dominated != 0);
    if (__pyx_t_2) {

      /* "gtfspy/routing/label.pyx":398
 *                     is_dominated = True
 *         if is_dominated:
 *             dominated.append(candidate)             # <<<<<<<<<<<<<<
 *         else:
 *             pareto_front.append(candidate)
 */
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_dominated, __pyx_v_candidate); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 398, __pyx_L1_error)

      /* "gtfspy/routing/label.pyx":397
 *                 if other.dominates(candidate):
 *                     is_dominated = True
 *         if is_dominated:             # <<<<<<<<<<<<<<
 *             dominated.append(candidate)
 *         else:
 */
      goto __pyx_L9;
    }

    /* "gtfspy/routing/label.pyx":400
 *             dominated.append(candidate)
 *         else:
 *             pareto_front.append(candidate)             # <<<<<<<<<<<<<<
 *         remaining = new_remaining
 *         # after each round:
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_pareto_front, __pyx_v_candidate); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 400, __pyx_L1_error)
    }
    __pyx_L9:;

    /* "gtfspy/routing/label.pyx":401
 *         else:
 *             pareto_front.append(candidate)
 *         remaining = new_remaining             # <<<<<<<<<<<<<<
 *         # after each round:
 *         #   remaining contains Labels that are not dominated by any in the pareto_front
 */
    __Pyx_INCREF(__pyx_v_new_remaining);
    __Pyx_DECREF_SET(__pyx_v_remaining, __pyx_v_new_remaining);
  }

  /* "gtfspy/routing/label.pyx":405
 *         #   remaining contains Labels that are not dominated by any in the pareto_front
 *         #   dominated contains Labels that are dominated by some other LabelTime
 *     return pareto_front             # <<<<<<<<<<<<<<
 * 
 * def merge_pareto_frontiers(labels, labels_other):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pareto_front);
  __pyx_r = __pyx_v_pareto_front;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":357
 *     return pareto_front
 * 
 * def compute_pareto_front_naive(list label_list):             # <<<<<<<<<<<<<<
 *     """
 *     Computes the Pareto frontier of a given label_list
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gtfspy.routing.label.compute_pareto_front_naive", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dominated);
  __Pyx_XDECREF(__pyx_v_pareto_front);
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_XDECREF(__pyx_v_candidate);
  __Pyx_XDECREF(__pyx_v_new_remaining);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":407
 *     return pareto_front
 * 
 * def merge_pareto_frontiers(labels, labels_other):             # <<<<<<<<<<<<<<
 *     """
 *     Merge two pareto frontiers by removing dominated entries.
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_7merge_pareto_frontiers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_6merge_pareto_frontiers[] = "\n    Merge two pareto frontiers by removing dominated entries.\n\n    Parameters\n    ----------\n    labels: list[LabelTime]\n    labels_other: list[LabelTime]\n\n    Returns\n    -------\n    pareto_front_merged: list[LabelTime]\n    ";
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_7merge_pareto_frontiers = {"merge_pareto_frontiers", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_7merge_pareto_frontiers, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6gtfspy_7routing_5label_6merge_pareto_frontiers};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_7merge_pareto_frontiers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_labels = 0;
  PyObject *__pyx_v_labels_other = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge_pareto_frontiers (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_labels,&__pyx_n_s_labels_other,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_labels)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_labels_other)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("merge_pareto_frontiers", 1, 2, 2, 1); __PYX_ERR(0, 407, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "merge_pareto_frontiers") < 0)) __PYX_ERR(0, 407, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_labels = values[0];
    __pyx_v_labels_other = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("merge_pareto_frontiers", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 407, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.merge_pareto_frontiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_6merge_pareto_frontiers(__pyx_self, __pyx_v_labels, __pyx_v_labels_other);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":420
 *     pareto_front_merged: list[LabelTime]
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):             # <<<<<<<<<<<<<<
 *         if survivor_list is None:
 *             survivor_list = list()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_22merge_pareto_frontiers_1_get_non_dominated_entries(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_22merge_pareto_frontiers_1_get_non_dominated_entries = {"_get_non_dominated_entries", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_22merge_pareto_frontiers_1_get_non_dominated_entries, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_22merge_pareto_frontiers_1_get_non_dominated_entries(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_candidates = 0;
  PyObject *__pyx_v_possible_dominators = 0;
  PyObject *__pyx_v_survivor_list = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_non_dominated_entries (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_candidates,&__pyx_n_s_possible_dominators,&__pyx_n_s_survivor_list,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)((PyObject *)Py_None));
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_candidates)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_possible_dominators)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_get_non_dominated_entries", 0, 2, 3, 1); __PYX_ERR(0, 420, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_survivor_list);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_get_non_dominated_entries") < 0)) __PYX_ERR(0, 420, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_candidates = values[0];
    __pyx_v_possible_dominators = values[1];
    __pyx_v_survivor_list = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_get_non_dominated_entries", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 420, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.merge_pareto_frontiers._get_non_dominated_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_22merge_pareto_frontiers__get_non_dominated_entries(__pyx_self, __pyx_v_candidates, __pyx_v_possible_dominators, __pyx_v_survivor_list);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_22merge_pareto_frontiers__get_non_dominated_entries(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_candidates, PyObject *__pyx_v_possible_dominators, PyObject *__pyx_v_survivor_list) {
  PyObject *__pyx_v_candidate = NULL;
  int __pyx_v_candidate_is_dominated;
  PyObject *__pyx_v_dominator = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("_get_non_dominated_entries", 0);
  __Pyx_INCREF(__pyx_v_survivor_list);

  /* "gtfspy/routing/label.pyx":421
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):
 *         if survivor_list is None:             # <<<<<<<<<<<<<<
 *             survivor_list = list()
 *         for candidate in candidates:
 */
  __pyx_t_1 = (__pyx_v_survivor_list == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "gtfspy/routing/label.pyx":422
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):
 *         if survivor_list is None:
 *             survivor_list = list()             # <<<<<<<<<<<<<<
 *         for candidate in candidates:
 *             candidate_is_dominated = False
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_survivor_list, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "gtfspy/routing/label.pyx":421
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):
 *         if survivor_list is None:             # <<<<<<<<<<<<<<
 *             survivor_list = list()
 *         for candidate in candidates:
 */
  }

  /* "gtfspy/routing/label.pyx":423
 *         if survivor_list is None:
 *             survivor_list = list()
 *         for candidate in candidates:             # <<<<<<<<<<<<<<
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:
 */
  if (likely(PyList_CheckExact(__pyx_v_candidates)) || PyTuple_CheckExact(__pyx_v_candidates)) {
    __pyx_t_3 = __pyx_v_candidates; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_candidates); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 423, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 423, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      }
    } else {
      __pyx_t_6 = __pyx_t_5(__pyx_t_3);
      if (unlikely(!__pyx_t_6)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 423, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_6);
    }
    __Pyx_XDECREF_SET(__pyx_v_candidate, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "gtfspy/routing/label.pyx":424
 *             survivor_list = list()
 *         for candidate in candidates:
 *             candidate_is_dominated = False             # <<<<<<<<<<<<<<
 *             for dominator in possible_dominators:
 *                 if dominator.dominates(candidate):
 */
    __pyx_v_candidate_is_dominated = 0;

    /* "gtfspy/routing/label.pyx":425
 *         for candidate in candidates:
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:             # <<<<<<<<<<<<<<
 *                 if dominator.dominates(candidate):
 *                     candidate_is_dominated = True
 */
    if (likely(PyList_CheckExact(__pyx_v_possible_dominators)) || PyTuple_CheckExact(__pyx_v_possible_dominators)) {
      __pyx_t_6 = __pyx_v_possible_dominators; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_possible_dominators); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 425, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 425, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 425, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 425, __pyx_L1_error)
          #else
          __pyx_t_9 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 425, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          #endif
        }
      } else {
        __pyx_t_9 = __pyx_t_8(__pyx_t_6);
        if (unlikely(!__pyx_t_9)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 425, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_9);
      }
      __Pyx_XDECREF_SET(__pyx_v_dominator, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "gtfspy/routing/label.pyx":426
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:
 *                 if dominator.dominates(candidate):             # <<<<<<<<<<<<<<
 *                     candidate_is_dominated = True
 *                     break
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_dominator, __pyx_n_s_dominates); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      if (!__pyx_t_11) {
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_candidate); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[2] = {__pyx_t_11, __pyx_v_candidate};
          __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
          PyObject *__pyx_temp[2] = {__pyx_t_11, __pyx_v_candidate};
          __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_9);
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(1+1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 426, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
          __Pyx_INCREF(__pyx_v_candidate);
          __Pyx_GIVEREF(__pyx_v_candidate);
          PyTuple_SET_ITEM(__pyx_t_12, 0+1, __pyx_v_candidate);
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_2) {

        /* "gtfspy/routing/label.pyx":427
 *             for dominator in possible_dominators:
 *                 if dominator.dominates(candidate):
 *                     candidate_is_dominated = True             # <<<<<<<<<<<<<<
 *                     break
 *             if not candidate_is_dominated:
 */
        __pyx_v_candidate_is_dominated = 1;

        /* "gtfspy/routing/label.pyx":428
 *                 if dominator.dominates(candidate):
 *                     candidate_is_dominated = True
 *                     break             # <<<<<<<<<<<<<<
 *             if not candidate_is_dominated:
 *                 survivor_list.append(candidate)
 */
        goto __pyx_L7_break;

        /* "gtfspy/routing/label.pyx":426
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:
 *                 if dominator.dominates(candidate):             # <<<<<<<<<<<<<<
 *                     candidate_is_dominated = True
 *                     break
 */
      }

      /* "gtfspy/routing/label.pyx":425
 *         for candidate in candidates:
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:             # <<<<<<<<<<<<<<
 *                 if dominator.dominates(candidate):
 *                     candidate_is_dominated = True
 */
    }
    __pyx_L7_break:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "gtfspy/routing/label.pyx":429
 *                     candidate_is_dominated = True
 *                     break
 *             if not candidate_is_dominated:             # <<<<<<<<<<<<<<
 *                 survivor_list.append(candidate)
 *         return survivor_list
 */
    __pyx_t_2 = ((!(__pyx_v_candidate_is_dominated != 0)) != 0);
    if (__pyx_t_2) {

      /* "gtfspy/routing/label.pyx":430
 *                     break
 *             if not candidate_is_dominated:
 *                 survivor_list.append(candidate)             # <<<<<<<<<<<<<<
 *         return survivor_list
 * 
 */
      __pyx_t_13 = __Pyx_PyObject_Append(__pyx_v_survivor_list, __pyx_v_candidate); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 430, __pyx_L1_error)

      /* "gtfspy/routing/label.pyx":429
 *                     candidate_is_dominated = True
 *                     break
 *             if not candidate_is_dominated:             # <<<<<<<<<<<<<<
 *                 survivor_list.append(candidate)
 *         return survivor_list
 */
    }

    /* "gtfspy/routing/label.pyx":423
 *         if survivor_list is None:
 *             survivor_list = list()
 *         for candidate in candidates:             # <<<<<<<<<<<<<<
 *             candidate_is_dominated = False
 *             for dominator in possible_dominators:
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":431
 *             if not candidate_is_dominated:
 *                 survivor_list.append(candidate)
 *         return survivor_list             # <<<<<<<<<<<<<<
 * 
 *     survived = []
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_survivor_list);
  __pyx_r = __pyx_v_survivor_list;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":420
 *     pareto_front_merged: list[LabelTime]
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):             # <<<<<<<<<<<<<<
 *         if survivor_list is None:
 *             survivor_list = list()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("gtfspy.routing.label.merge_pareto_frontiers._get_non_dominated_entries", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candidate);
  __Pyx_XDECREF(__pyx_v_dominator);
  __Pyx_XDECREF(__pyx_v_survivor_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":407
 *     return pareto_front
 * 
 * def merge_pareto_frontiers(labels, labels_other):             # <<<<<<<<<<<<<<
 *     """
 *     Merge two pareto frontiers by removing dominated entries.
 */

static PyObject *__pyx_pf_6gtfspy_7routing_5label_6merge_pareto_frontiers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_labels, PyObject *__pyx_v_labels_other) {
  PyObject *__pyx_v__get_non_dominated_entries = 0;
  PyObject *__pyx_v_survived = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("merge_pareto_frontiers", 0);

  /* "gtfspy/routing/label.pyx":420
 *     pareto_front_merged: list[LabelTime]
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):             # <<<<<<<<<<<<<<
 *         if survivor_list is None:
 *             survivor_list = list()
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_22merge_pareto_frontiers_1_get_non_dominated_entries, 0, __pyx_n_s_merge_pareto_frontiers_locals__g, NULL, __pyx_n_s_gtfspy_routing_label, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_1, __pyx_tuple__8);
  __pyx_v__get_non_dominated_entries = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":433
 *         return survivor_list
 * 
 *     survived = []             # <<<<<<<<<<<<<<
 *     survived = _get_non_dominated_entries(labels, labels_other, survivor_list=survived)
 *     survived = _get_non_dominated_entries(labels_other, survived, survivor_list=survived)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_survived = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":434
 * 
 *     survived = []
 *     survived = _get_non_dominated_entries(labels, labels_other, survivor_list=survived)             # <<<<<<<<<<<<<<
 *     survived = _get_non_dominated_entries(labels_other, survived, survivor_list=survived)
 *     return survived
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_labels);
  __Pyx_GIVEREF(__pyx_v_labels);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_labels);
  __Pyx_INCREF(__pyx_v_labels_other);
  __Pyx_GIVEREF(__pyx_v_labels_other);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_labels_other);
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_survivor_list, __pyx_v_survived) < 0) __PYX_ERR(0, 434, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_v__get_non_dominated_entries, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_survived, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":435
 *     survived = []
 *     survived = _get_non_dominated_entries(labels, labels_other, survivor_list=survived)
 *     survived = _get_non_dominated_entries(labels_other, survived, survivor_list=survived)             # <<<<<<<<<<<<<<
 *     return survived
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_labels_other);
  __Pyx_GIVEREF(__pyx_v_labels_other);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_labels_other);
  __Pyx_INCREF(__pyx_v_survived);
  __Pyx_GIVEREF(__pyx_v_survived);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_survived);
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_survivor_list, __pyx_v_survived) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v__get_non_dominated_entries, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_survived, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":436
 *     survived = _get_non_dominated_entries(labels, labels_other, survivor_list=survived)
 *     survived = _get_non_dominated_entries(labels_other, survived, survivor_list=survived)
 *     return survived             # <<<<<<<<<<<<<<
 * 
 * def min_arrival_time_target(label_list):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_survived);
  __pyx_r = __pyx_v_survived;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":407
 *     return pareto_front
 * 
 * def merge_pareto_frontiers(labels, labels_other):             # <<<<<<<<<<<<<<
 *     """
 *     Merge two pareto frontiers by removing dominated entries.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.merge_pareto_frontiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__get_non_dominated_entries);
  __Pyx_XDECREF(__pyx_v_survived);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":438
 *     return survived
 * 
 * def min_arrival_time_target(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9min_arrival_time_target(PyObject *__pyx_self, PyObject *__pyx_v_label_list); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_9min_arrival_time_target = {"min_arrival_time_target", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9min_arrival_time_target, METH_O, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_9min_arrival_time_target(PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("min_arrival_time_target (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_8min_arrival_time_target(__pyx_self, ((PyObject *)__pyx_v_label_list));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":440
 * def min_arrival_time_target(label_list):
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target             # <<<<<<<<<<<<<<
 *     else:
 *         return float('inf')
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_23min_arrival_time_target_lambda(PyObject *__pyx_self, PyObject *__pyx_v_label); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_23min_arrival_time_target_lambda = {"lambda", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_23min_arrival_time_target_lambda, METH_O, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_23min_arrival_time_target_lambda(PyObject *__pyx_self, PyObject *__pyx_v_label) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, ((PyObject *)__pyx_v_label));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.min_arrival_time_target.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":438
 *     return survived
 * 
 * def min_arrival_time_target(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 */

static PyObject *__pyx_pf_6gtfspy_7routing_5label_8min_arrival_time_target(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("min_arrival_time_target", 0);

  /* "gtfspy/routing/label.pyx":439
 * 
 * def min_arrival_time_target(label_list):
 *     if len(label_list) > 0:             # <<<<<<<<<<<<<<
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 *     else:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_label_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 439, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 > 0) != 0);
  if (__pyx_t_2) {

    /* "gtfspy/routing/label.pyx":440
 * def min_arrival_time_target(label_list):
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target             # <<<<<<<<<<<<<<
 *     else:
 *         return float('inf')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_label_list);
    __Pyx_GIVEREF(__pyx_v_label_list);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_label_list);
    __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_23min_arrival_time_target_lambda, 0, __pyx_n_s_min_arrival_time_target_locals_l, NULL, __pyx_n_s_gtfspy_routing_label, __pyx_d, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_arrival_time_target); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":439
 * 
 * def min_arrival_time_target(label_list):
 *     if len(label_list) > 0:             # <<<<<<<<<<<<<<
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 *     else:
 */
  }

  /* "gtfspy/routing/label.pyx":442
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 *     else:
 *         return float('inf')             # <<<<<<<<<<<<<<
 * 
 * def min_n_vehicle_trips(label_list):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_n_s_inf); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":438
 *     return survived
 * 
 * def min_arrival_time_target(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.min_arrival_time_target", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":444
 *         return float('inf')
 * 
 * def min_n_vehicle_trips(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_11min_n_vehicle_trips(PyObject *__pyx_self, PyObject *__pyx_v_label_list); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_11min_n_vehicle_trips = {"min_n_vehicle_trips", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_11min_n_vehicle_trips, METH_O, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_11min_n_vehicle_trips(PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("min_n_vehicle_trips (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_10min_n_vehicle_trips(__pyx_self, ((PyObject *)__pyx_v_label_list));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":446
 * def min_n_vehicle_trips(label_list):
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips             # <<<<<<<<<<<<<<
 *     else:
 *         return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_19min_n_vehicle_trips_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_label); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_19min_n_vehicle_trips_lambda1 = {"lambda1", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_19min_n_vehicle_trips_lambda1, METH_O, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_19min_n_vehicle_trips_lambda1(PyObject *__pyx_self, PyObject *__pyx_v_label) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self, ((PyObject *)__pyx_v_label));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_label, __pyx_n_s_n_vehicle_trips); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.min_n_vehicle_trips.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":444
 *         return float('inf')
 * 
 * def min_n_vehicle_trips(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 */

static PyObject *__pyx_pf_6gtfspy_7routing_5label_10min_n_vehicle_trips(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_label_list) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("min_n_vehicle_trips", 0);

  /* "gtfspy/routing/label.pyx":445
 * 
 * def min_n_vehicle_trips(label_list):
 *     if len(label_list) > 0:             # <<<<<<<<<<<<<<
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 *     else:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_label_list); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 > 0) != 0);
  if (__pyx_t_2) {

    /* "gtfspy/routing/label.pyx":446
 * def min_n_vehicle_trips(label_list):
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips             # <<<<<<<<<<<<<<
 *     else:
 *         return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_label_list);
    __Pyx_GIVEREF(__pyx_v_label_list);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_label_list);
    __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_19min_n_vehicle_trips_lambda1, 0, __pyx_n_s_min_n_vehicle_trips_locals_lambd, NULL, __pyx_n_s_gtfspy_routing_label, __pyx_d, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_n_vehicle_trips); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":445
 * 
 * def min_n_vehicle_trips(label_list):
 *     if len(label_list) > 0:             # <<<<<<<<<<<<<<
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 *     else:
 */
  }

  /* "gtfspy/routing/label.pyx":448
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 *     else:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * cdef class LabelTimeBoardingsAndRoute:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":444
 *         return float('inf')
 * 
 * def min_n_vehicle_trips(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.min_n_vehicle_trips", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":463
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_arrival_time_target;
  int __pyx_v_n_boardings;
  int __pyx_v_movement_duration;
  int __pyx_v_first_leg_is_walk;
  PyObject *__pyx_v_connection = 0;
  PyObject *__pyx_v_previous_label = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_arrival_time_target,&__pyx_n_s_n_boardings,&__pyx_n_s_movement_duration,&__pyx_n_s_first_leg_is_walk,&__pyx_n_s_connection,&__pyx_n_s_previous_label,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};

    /* "gtfspy/routing/label.pyx":464
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 */
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrival_time_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 1); __PYX_ERR(0, 463, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_n_boardings)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 2); __PYX_ERR(0, 463, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_movement_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 3); __PYX_ERR(0, 463, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_first_leg_is_walk)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, 4); __PYX_ERR(0, 463, __pyx_L3_error)
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connection);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_previous_label);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 463, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L3_error)
    __pyx_v_arrival_time_target = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_arrival_time_target == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L3_error)
    __pyx_v_n_boardings = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n_boardings == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_movement_duration = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_movement_duration == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_first_leg_is_walk = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_first_leg_is_walk == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L3_error)
    __pyx_v_connection = values[5];
    __pyx_v_previous_label = values[6];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 5, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 463, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), __pyx_v_departure_time, __pyx_v_arrival_time_target, __pyx_v_n_boardings, __pyx_v_movement_duration, __pyx_v_first_leg_is_walk, __pyx_v_connection, __pyx_v_previous_label);

  /* "gtfspy/routing/label.pyx":463
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_n_boardings, int __pyx_v_movement_duration, int __pyx_v_first_leg_is_walk, PyObject *__pyx_v_connection, PyObject *__pyx_v_previous_label) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":465
 *     def __init__(self, double departure_time, double arrival_time_target,
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":466
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target             # <<<<<<<<<<<<<<
 *         self.n_boardings = n_boardings
 *         self.movement_duration = movement_duration
 */
  __pyx_v_self->arrival_time_target = __pyx_v_arrival_time_target;

  /* "gtfspy/routing/label.pyx":467
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings             # <<<<<<<<<<<<<<
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk
 */
  __pyx_v_self->n_boardings = __pyx_v_n_boardings;

  /* "gtfspy/routing/label.pyx":468
 *         self.arrival_time_target = arrival_time_target
 *         self.n_boardings = n_boardings
 *         self.movement_duration = movement_duration             # <<<<<<<<<<<<<<
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label
 */
  __pyx_v_self->movement_duration = __pyx_v_movement_duration;

  /* "gtfspy/routing/label.pyx":469
 *         self.n_boardings = n_boardings
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk             # <<<<<<<<<<<<<<
 *         self.previous_label = previous_label
 *         self.connection = connection
 */
  __pyx_v_self->first_leg_is_walk = __pyx_v_first_leg_is_walk;

  /* "gtfspy/routing/label.pyx":470
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label             # <<<<<<<<<<<<<<
 *         self.connection = connection
 * 
 */
  __Pyx_INCREF(__pyx_v_previous_label);
  __Pyx_GIVEREF(__pyx_v_previous_label);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_v_previous_label;

  /* "gtfspy/routing/label.pyx":471
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label
 *         self.connection = connection             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_INCREF(__pyx_v_connection);
  __Pyx_GIVEREF(__pyx_v_connection);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_v_connection;

  /* "gtfspy/routing/label.pyx":463
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int n_boardings, int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":473
 *         self.connection = connection
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_2__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":474
 * 
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->connection);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __Pyx_GIVEREF(__pyx_v_self->previous_label);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->previous_label);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":473
 *         self.connection = connection
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":476
 *         return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_4__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  double __pyx_t_9;
  double __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":477
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state             # <<<<<<<<<<<<<<
 * 
 *     def _tuple_for_ordering(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 6)) {
      if (size > 6) __Pyx_RaiseTooManyValuesError(6);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 477, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 5); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 5); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[6] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
      for (i=0; i < 6; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 477, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[6] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6};
    __pyx_t_7 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
    for (index=0; index < 6; index++) {
      PyObject* item = __pyx_t_8(__pyx_t_7); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 6) < 0) __PYX_ERR(0, 477, __pyx_L1_error)
    __pyx_t_8 = NULL;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 477, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_10 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_self->departure_time = __pyx_t_9;
  __pyx_v_self->arrival_time_target = __pyx_t_10;
  __pyx_v_self->n_boardings = __pyx_t_11;
  __pyx_v_self->movement_duration = __pyx_t_12;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_t_5;
  __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_6);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "gtfspy/routing/label.pyx":476
 *         return self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":479
 *         self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_6_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);

  /* "gtfspy/routing/label.pyx":480
 * 
 *     def _tuple_for_ordering(self):
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int((-__pyx_v_self->n_boardings)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong((!(__pyx_v_self->first_leg_is_walk != 0))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int((-__pyx_v_self->movement_duration)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":479
 *         self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":482
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration
 * 
 *     def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "self", 0))) __PYX_ERR(0, 482, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 482, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_8__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":483
 * 
 *     def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":484
 *     def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 484, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":485
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":486
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":485
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":487
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":488
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":487
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":489
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":490
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":489
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":491
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":492
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":491
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":493
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":494
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":493
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":495
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":496
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelTimeBoardingsAndRoute other):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":495
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":482
 *         return self.departure_time, -self.arrival_time_target, -self.n_boardings, not self.first_leg_is_walk, -self.movement_duration
 * 
 *     def __richcmp__(LabelTimeBoardingsAndRoute self, LabelTimeBoardingsAndRoute other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":498
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelTimeBoardingsAndRoute dominates the other LabelTimeBoardingsAndRoute
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 498, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 498, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 498, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":509
 *             True if this ParetoTuple dominates the other, otherwise False
 *         """
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":510
 *         """
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 510, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":511
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__9, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__10, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 511, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 511, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 511, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 511, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 511, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 511, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 511, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 511, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":512
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":511
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  }

  /* "gtfspy/routing/label.pyx":513
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__11, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__12, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 513, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 513, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 513, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 513, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 513, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 513, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 513, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_unpacking_done;
      __pyx_L12_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 513, __pyx_L1_error)
      __pyx_L13_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_other_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_11 = __pyx_t_12;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":514
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":513
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "gtfspy/routing/label.pyx":516
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":498
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelTimeBoardingsAndRoute dominates the other LabelTimeBoardingsAndRoute
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10dominates[] = "\n        Compute whether this LabelTimeBoardingsAndRoute dominates the other LabelTimeBoardingsAndRoute\n        Parameters\n        ----------\n        other: LabelTimeBoardingsAndRoute\n        Returns\n        -------\n        dominates: bint\n            True if this ParetoTuple dominates the other, otherwise False\n        ";
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":518
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_13dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 518, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 518, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":519
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target > __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings > __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":520
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
 *                 and self.movement_duration > other.movement_duration:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":519
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
 */
  }

  /* "gtfspy/routing/label.pyx":521
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \             # <<<<<<<<<<<<<<
 *                 and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target == __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings == __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":522
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
 *                 and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_8 = ((__pyx_v_self->movement_duration > __pyx_v_other->movement_duration) != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;

  /* "gtfspy/routing/label.pyx":521
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \             # <<<<<<<<<<<<<<
 *                 and self.movement_duration > other.movement_duration:
 *             return False
 */
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":523
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \
 *                 and self.movement_duration > other.movement_duration:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":521
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target, self.n_boardings == other.n_boardings]) \             # <<<<<<<<<<<<<<
 *                 and self.movement_duration > other.movement_duration:
 *             return False
 */
  }

  /* "gtfspy/routing/label.pyx":525
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":518
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target, self.n_boardings > other.n_boardings]):
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 518, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_12dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":527
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.n_boardings > other.n_boardings,
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_15dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 527, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 527, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":528
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target > __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":529
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.n_boardings > other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings > __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":530
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk > __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":528
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":531
 *                 self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.n_boardings == other.n_boardings,
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":528
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */
  }

  /* "gtfspy/routing/label.pyx":532
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
  __pyx_t_5 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target == __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "gtfspy/routing/label.pyx":533
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.n_boardings == other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings == __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":534
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk == __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":532
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":534
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_8 = ((__pyx_v_self->movement_duration > __pyx_v_other->movement_duration) != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;

  /* "gtfspy/routing/label.pyx":532
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":535
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":532
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
  }

  /* "gtfspy/routing/label.pyx":537
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":527
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.n_boardings > other.n_boardings,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 527, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":539
 *             return True
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17dominates_ignoring_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 539, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":540
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.n_boardings > other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings > __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":541
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk > __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":540
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.n_boardings > other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":542
 *         if any([self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":540
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.n_boardings > other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  }

  /* "gtfspy/routing/label.pyx":543
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_self->n_boardings == __pyx_v_other->n_boardings)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":544
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk == __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":543
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":544
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_8 = ((__pyx_v_self->movement_duration > __pyx_v_other->movement_duration) != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;

  /* "gtfspy/routing/label.pyx":543
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":545
 *         elif all([self.n_boardings == other.n_boardings,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":543
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.n_boardings == other.n_boardings,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  }

  /* "gtfspy/routing/label.pyx":547
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":539
 *             return True
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.n_boardings > other.n_boardings,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 539, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_16dominates_ignoring_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":549
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_and); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19dominates_ignoring_dep_time_and_n_boardings)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 549, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":550
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target > __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":551
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk > __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":550
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":552
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":550
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):
 *         if any([self.arrival_time_target > other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 */
  }

  /* "gtfspy/routing/label.pyx":553
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target == __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":554
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_v_self->first_leg_is_walk == __pyx_v_other->first_leg_is_walk)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":553
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":554
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_8 = ((__pyx_v_self->movement_duration > __pyx_v_other->movement_duration) != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;

  /* "gtfspy/routing/label.pyx":553
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  if (__pyx_t_7) {

    /* "gtfspy/routing/label.pyx":555
 *         elif all([self.arrival_time_target == other.arrival_time_target,
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":553
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 *             return False
 *         elif all([self.arrival_time_target == other.arrival_time_target,             # <<<<<<<<<<<<<<
 *                 self.first_leg_is_walk == other.first_leg_is_walk]) and self.movement_duration > other.movement_duration:
 *             return False
 */
  }

  /* "gtfspy/routing/label.pyx":557
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":549
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeBoardingsAndRoute other):             # <<<<<<<<<<<<<<
 *         if any([self.arrival_time_target > other.arrival_time_target,
 *                 self.first_leg_is_walk > other.first_leg_is_walk]):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, 1, "other", 0))) __PYX_ERR(0, 549, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_18dominates_ignoring_dep_time_and_n_boardings(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_and_n_boardings(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":559
 *             return True
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_label_with_connection_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_21get_label_with_connection_added)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_connection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_connection);
          __Pyx_GIVEREF(__pyx_v_connection);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_connection);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":560
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":561
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":560
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":561
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 561, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 561, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":560
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), __pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":559
 *             return True
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_label_with_connection_added (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_20get_label_with_connection_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_connection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_label_with_connection_added(__pyx_v_self, __pyx_v_connection, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":562
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_23get_copy)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":563
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":564
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "gtfspy/routing/label.pyx":563
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_6 = PyTuple_New(6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":564
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_previous_label, __pyx_v_self->previous_label) < 0) __PYX_ERR(0, 564, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":563
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":562
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_22get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":566
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_specified_departur); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_25get_copy_with_specified_departure_time)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_departure_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_departure_time);
          __Pyx_GIVEREF(__pyx_v_departure_time);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_departure_time);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":567
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":568
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "gtfspy/routing/label.pyx":567
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_4 = PyTuple_New(6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":568
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_previous_label, __pyx_v_self->previous_label) < 0) __PYX_ERR(0, 568, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":567
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":566
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, self.arrival_time_target,
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_24get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_specified_departure_time(__pyx_v_self, __pyx_v_departure_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":570
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_27duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 570, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":571
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":570
 *                                            self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeBoardingsAndRoute.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_26duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":574
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_departure_time = 0;
  PyObject *__pyx_v_walk_duration = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 574, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 574, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = values[0];
    __pyx_v_walk_duration = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 574, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_28direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":575
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_departure_time, __pyx_v_walk_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_0);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":574
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":577
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_walk_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_31get_copy_with_walk_added)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_walk_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 577, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_connection);
        __Pyx_GIVEREF(__pyx_v_connection);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_connection);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute))))) __PYX_ERR(0, 577, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":578
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->departure_time - __pyx_v_walk_duration)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":579
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyFloat_FromDouble((__pyx_v_self->movement_duration + __pyx_v_walk_duration)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "gtfspy/routing/label.pyx":578
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_7);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_3, 4, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;
  __pyx_t_7 = 0;

  /* "gtfspy/routing/label.pyx":579
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 579, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":578
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), __pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":577
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeBoardingsAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_walk_duration;
  PyObject *__pyx_v_connection = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_walk_duration,&__pyx_n_s_connection,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connection)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, 1); __PYX_ERR(0, 577, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_copy_with_walk_added") < 0)) __PYX_ERR(0, 577, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 577, __pyx_L3_error)
    __pyx_v_connection = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 577, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_30get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), __pyx_v_walk_duration, __pyx_v_connection);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_walk_added(__pyx_v_self, __pyx_v_walk_duration, __pyx_v_connection, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":581
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_33__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_33__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_32__str__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gtfspy/routing/label.pyx":582
 * 
 *     def __str__(self):
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))             # <<<<<<<<<<<<<<
 * 
 * cdef class LabelTimeAndRoute:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __Pyx_GIVEREF(__pyx_v_self->previous_label);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_v_self->previous_label);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":581
 *                                            self.arrival_time_target, self.n_boardings, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.n_boardings, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":454
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public double arrival_time_target
 *         public int n_boardings
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 454, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":455
 *     cdef:
 *         public double departure_time
 *         public double arrival_time_target             # <<<<<<<<<<<<<<
 *         public int n_boardings
 *         public int movement_duration
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 455, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":456
 *         public double departure_time
 *         public double arrival_time_target
 *         public int n_boardings             # <<<<<<<<<<<<<<
 *         public int movement_duration
 *         public bint first_leg_is_walk
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.n_boardings.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 456, __pyx_L1_error)
  __pyx_v_self->n_boardings = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.n_boardings.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":457
 *         public double arrival_time_target
 *         public int n_boardings
 *         public int movement_duration             # <<<<<<<<<<<<<<
 *         public bint first_leg_is_walk
 *         public object previous_label
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.movement_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_v_self->movement_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.movement_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":458
 *         public int n_boardings
 *         public int movement_duration
 *         public bint first_leg_is_walk             # <<<<<<<<<<<<<<
 *         public object previous_label
 *         public object connection
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.first_leg_is_walk.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 458, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeBoardingsAndRoute.first_leg_is_walk.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":459
 *         public int movement_duration
 *         public bint first_leg_is_walk
 *         public object previous_label             # <<<<<<<<<<<<<<
 *         public object connection
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __pyx_r = __pyx_v_self->previous_label;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_4__del__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":460
 *         public bint first_leg_is_walk
 *         public object previous_label
 *         public object connection             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->connection);
  __pyx_r = __pyx_v_self->connection;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_4__del__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":596
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_departure_time;
  double __pyx_v_arrival_time_target;
  int __pyx_v_movement_duration;
  int __pyx_v_first_leg_is_walk;
  PyObject *__pyx_v_connection = 0;
  PyObject *__pyx_v_previous_label = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_arrival_time_target,&__pyx_n_s_movement_duration,&__pyx_n_s_first_leg_is_walk,&__pyx_n_s_connection,&__pyx_n_s_previous_label,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "gtfspy/routing/label.pyx":597
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_arrival_time_target)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 1); __PYX_ERR(0, 596, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_movement_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 2); __PYX_ERR(0, 596, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_first_leg_is_walk)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 3); __PYX_ERR(0, 596, __pyx_L3_error)
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connection);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_previous_label);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 596, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_departure_time = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_departure_time == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L3_error)
    __pyx_v_arrival_time_target = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_arrival_time_target == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L3_error)
    __pyx_v_movement_duration = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_movement_duration == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 597, __pyx_L3_error)
    __pyx_v_first_leg_is_walk = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_first_leg_is_walk == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 597, __pyx_L3_error)
    __pyx_v_connection = values[4];
    __pyx_v_previous_label = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 596, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), __pyx_v_departure_time, __pyx_v_arrival_time_target, __pyx_v_movement_duration, __pyx_v_first_leg_is_walk, __pyx_v_connection, __pyx_v_previous_label);

  /* "gtfspy/routing/label.pyx":596
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_departure_time, double __pyx_v_arrival_time_target, int __pyx_v_movement_duration, int __pyx_v_first_leg_is_walk, PyObject *__pyx_v_connection, PyObject *__pyx_v_previous_label) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":598
 *     def __init__(self, double departure_time, double arrival_time_target,
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time             # <<<<<<<<<<<<<<
 *         self.arrival_time_target = arrival_time_target
 *         self.movement_duration = movement_duration
 */
  __pyx_v_self->departure_time = __pyx_v_departure_time;

  /* "gtfspy/routing/label.pyx":599
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target             # <<<<<<<<<<<<<<
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk
 */
  __pyx_v_self->arrival_time_target = __pyx_v_arrival_time_target;

  /* "gtfspy/routing/label.pyx":600
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 *         self.movement_duration = movement_duration             # <<<<<<<<<<<<<<
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label
 */
  __pyx_v_self->movement_duration = __pyx_v_movement_duration;

  /* "gtfspy/routing/label.pyx":601
 *         self.arrival_time_target = arrival_time_target
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk             # <<<<<<<<<<<<<<
 *         self.previous_label = previous_label
 *         self.connection = connection
 */
  __pyx_v_self->first_leg_is_walk = __pyx_v_first_leg_is_walk;

  /* "gtfspy/routing/label.pyx":602
 *         self.movement_duration = movement_duration
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label             # <<<<<<<<<<<<<<
 *         self.connection = connection
 * 
 */
  __Pyx_INCREF(__pyx_v_previous_label);
  __Pyx_GIVEREF(__pyx_v_previous_label);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_v_previous_label;

  /* "gtfspy/routing/label.pyx":603
 *         self.first_leg_is_walk = first_leg_is_walk
 *         self.previous_label = previous_label
 *         self.connection = connection             # <<<<<<<<<<<<<<
 * 
 *     def __getstate__(self):
 */
  __Pyx_INCREF(__pyx_v_connection);
  __Pyx_GIVEREF(__pyx_v_connection);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_v_connection;

  /* "gtfspy/routing/label.pyx":596
 * 
 * 
 *     def __init__(self, double departure_time, double arrival_time_target,             # <<<<<<<<<<<<<<
 *                  int movement_duration, bint first_leg_is_walk, object connection=None, object previous_label=None):
 *         self.departure_time = departure_time
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":605
 *         self.connection = connection
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_2__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":606
 * 
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_self->connection);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __Pyx_GIVEREF(__pyx_v_self->previous_label);
  PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_self->previous_label);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":605
 *         self.connection = connection
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":608
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_4__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":609
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state             # <<<<<<<<<<<<<<
 * 
 *     def _tuple_for_ordering(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 5)) {
      if (size > 5) __Pyx_RaiseTooManyValuesError(5);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 609, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 5; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 609, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 5; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 5) < 0) __PYX_ERR(0, 609, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 609, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->departure_time = __pyx_t_8;
  __pyx_v_self->arrival_time_target = __pyx_t_9;
  __pyx_v_self->movement_duration = __pyx_t_10;
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_t_4;
  __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":608
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":611
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_6_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);

  /* "gtfspy/routing/label.pyx":612
 * 
 *     def _tuple_for_ordering(self):
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(__pyx_v_self->first_leg_is_walk != 0))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int((-__pyx_v_self->movement_duration)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":611
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":614
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration
 * 
 *     def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "self", 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_8__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":615
 * 
 *     def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":616
 *     def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":617
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":618
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":617
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":619
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":620
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":619
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":621
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":622
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":621
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":623
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":624
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":623
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":625
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":626
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":625
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":627
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":628
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelTimeAndRoute other):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":627
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":614
 *         return self.departure_time, -self.arrival_time_target, not self.first_leg_is_walk, -self.movement_duration
 * 
 *     def __richcmp__(LabelTimeAndRoute self, LabelTimeAndRoute other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":630
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 630, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 630, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 630, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":641
 *             True if this ParetoTuple dominates the other, otherwise False
 *         """
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":642
 *         """
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":643
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__13, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__14, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 643, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 643, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 643, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 643, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 643, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 643, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":644
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":643
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  }

  /* "gtfspy/routing/label.pyx":645
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__15, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__16, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 645, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 645, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 645, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_unpacking_done;
      __pyx_L12_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 645, __pyx_L1_error)
      __pyx_L13_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 645, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_other_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_11 = __pyx_t_12;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":646
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":645
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "gtfspy/routing/label.pyx":648
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeAndRoute other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":630
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_17LabelTimeAndRoute_10dominates[] = "\n        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles\n        Parameters\n        ----------\n        other: LabelTimeAndRoute\n        Returns\n        -------\n        dominates: bint\n            True if this ParetoTuple dominates the other, otherwise False\n        ";
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 630, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":650
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_dominates = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_13dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 650, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 650, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":652
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeAndRoute other):
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dominates = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":654
 *             self.arrival_time_target <= other.arrival_time_target
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeAndRoute other):
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_dominates); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":650
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dominates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 650, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_12dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":656
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_15dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 656, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":660
 *             int dominates
 *         dominates = (
 *             (self.arrival_time_target <= other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk) or
 *             (self.arrival_time_target == other.arrival_time_target and
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (!__pyx_t_7) {
    goto __pyx_L4_next_or;
  } else {
  }

  /* "gtfspy/routing/label.pyx":661
 *         dominates = (
 *             (self.arrival_time_target <= other.arrival_time_target and
 *             self.first_leg_is_walk <= other.first_leg_is_walk) or             # <<<<<<<<<<<<<<
 *             (self.arrival_time_target == other.arrival_time_target and
 *             self.first_leg_is_walk == other.first_leg_is_walk and self.movement_duration <= other.movement_duration)
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_L4_next_or:;

  /* "gtfspy/routing/label.pyx":662
 *             (self.arrival_time_target <= other.arrival_time_target and
 *             self.first_leg_is_walk <= other.first_leg_is_walk) or
 *             (self.arrival_time_target == other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk == other.first_leg_is_walk and self.movement_duration <= other.movement_duration)
 * 
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target == __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":663
 *             self.first_leg_is_walk <= other.first_leg_is_walk) or
 *             (self.arrival_time_target == other.arrival_time_target and
 *             self.first_leg_is_walk == other.first_leg_is_walk and self.movement_duration <= other.movement_duration)             # <<<<<<<<<<<<<<
 * 
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk == __pyx_v_other->first_leg_is_walk);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_self->movement_duration <= __pyx_v_other->movement_duration);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":666
 * 
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeAndRoute other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":656
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":668
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17dominates_ignoring_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 668, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":672
 *             int dominates
 *         dominates = (
 *             self.movement_duration <= other.movement_duration and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->movement_duration <= __pyx_v_other->movement_duration);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":673
 *         dominates = (
 *             self.movement_duration <= other.movement_duration and
 *             self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":675
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeAndRoute other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":668
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 668, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_16dominates_ignoring_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":677
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_and); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19dominates_ignoring_dep_time_and_n_boardings)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 677, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":681
 *             int dominates
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and             # <<<<<<<<<<<<<<
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":682
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target and
 *             self.first_leg_is_walk <= other.first_leg_is_walk             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_7 = (__pyx_v_self->first_leg_is_walk <= __pyx_v_other->first_leg_is_walk);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":684
 *             self.first_leg_is_walk <= other.first_leg_is_walk
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":677
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelTimeAndRoute other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, 1, "other", 0))) __PYX_ERR(0, 677, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_18dominates_ignoring_dep_time_and_n_boardings(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_and_n_boardings(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 677, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":686
 *         return dominates
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_label_with_connection_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 686, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_21get_label_with_connection_added)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_connection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 686, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_connection);
          __Pyx_GIVEREF(__pyx_v_connection);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_connection);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 686, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":687
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":688
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "gtfspy/routing/label.pyx":687
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":688
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 688, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 688, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":687
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute), __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":686
 *         return dominates
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_label_with_connection_added (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_20get_label_with_connection_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_connection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_label_with_connection_added(__pyx_v_self, __pyx_v_connection, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":689
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_23get_copy)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 689, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":690
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":691
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":690
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":691
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_previous_label, __pyx_v_self->previous_label) < 0) __PYX_ERR(0, 691, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":690
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute), __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":689
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_22get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":693
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_specified_departur); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_25get_copy_with_specified_departure_time)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_departure_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_departure_time);
          __Pyx_GIVEREF(__pyx_v_departure_time);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_departure_time);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":694
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":695
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":694
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":695
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_previous_label, __pyx_v_self->previous_label) < 0) __PYX_ERR(0, 695, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":694
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute), __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":693
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_24get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_specified_departure_time(__pyx_v_self, __pyx_v_departure_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":697
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_27duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":698
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":697
 *                                            self.movement_duration, self.first_leg_is_walk, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelTimeAndRoute.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_26duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":701
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_departure_time = 0;
  PyObject *__pyx_v_walk_duration = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 701, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 701, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = values[0];
    __pyx_v_walk_duration = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 701, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_28direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":702
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_departure_time, __pyx_v_walk_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_0);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":701
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":704
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_walk_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_31get_copy_with_walk_added)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_walk_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_connection);
        __Pyx_GIVEREF(__pyx_v_connection);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_connection);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 704, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute))))) __PYX_ERR(0, 704, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":705
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->departure_time - __pyx_v_walk_duration)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":706
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->movement_duration + __pyx_v_walk_duration)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":705
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_7, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":706
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 706, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 706, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":705
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute), __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":704
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelTimeAndRoute get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_walk_duration;
  PyObject *__pyx_v_connection = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_walk_duration,&__pyx_n_s_connection,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connection)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, 1); __PYX_ERR(0, 704, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_copy_with_walk_added") < 0)) __PYX_ERR(0, 704, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 704, __pyx_L3_error)
    __pyx_v_connection = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 704, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_30get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), __pyx_v_walk_duration, __pyx_v_connection);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_walk_added(__pyx_v_self, __pyx_v_walk_duration, __pyx_v_connection, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":708
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_33__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_33__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_32__str__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gtfspy/routing/label.pyx":709
 * 
 *     def __str__(self):
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __Pyx_GIVEREF(__pyx_v_self->previous_label);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->previous_label);
  __Pyx_INCREF(__pyx_v_self->connection);
  __Pyx_GIVEREF(__pyx_v_self->connection);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->connection);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":708
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration, self.first_leg_is_walk, self.previous_label, self.connection))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":588
 *     # The trip with minimal "movement time" should be chosen = maximizing the waiting time for robustness
 *     cdef:
 *         public double departure_time             # <<<<<<<<<<<<<<
 *         public double arrival_time_target
 *         public int movement_duration
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":589
 *     cdef:
 *         public double departure_time
 *         public double arrival_time_target             # <<<<<<<<<<<<<<
 *         public int movement_duration
 *         public bint first_leg_is_walk
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 589, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":590
 *         public double departure_time
 *         public double arrival_time_target
 *         public int movement_duration             # <<<<<<<<<<<<<<
 *         public bint first_leg_is_walk
 *         public object previous_label
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.movement_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 590, __pyx_L1_error)
  __pyx_v_self->movement_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.movement_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":591
 *         public double arrival_time_target
 *         public int movement_duration
 *         public bint first_leg_is_walk             # <<<<<<<<<<<<<<
 *         public object previous_label
 *         public object connection
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->first_leg_is_walk); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.first_leg_is_walk.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 591, __pyx_L1_error)
  __pyx_v_self->first_leg_is_walk = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelTimeAndRoute.first_leg_is_walk.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":592
 *         public int movement_duration
 *         public bint first_leg_is_walk
 *         public object previous_label             # <<<<<<<<<<<<<<
 *         public object connection
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->previous_label);
  __pyx_r = __pyx_v_self->previous_label;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_4__del__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->previous_label);
  __Pyx_DECREF(__pyx_v_self->previous_label);
  __pyx_v_self->previous_label = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":593
 *         public bint first_leg_is_walk
 *         public object previous_label
 *         public object connection             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->connection);
  __pyx_r = __pyx_v_self->connection;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_4__del__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_4__del__(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->connection);
  __Pyx_DECREF(__pyx_v_self->connection);
  __pyx_v_self->connection = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":718
 *     cdef public double departure_time, arrival_time_target
 * 
 *     def __init__(self, journey_dict):             # <<<<<<<<<<<<<<
 *         for key in journey_dict:
 *             setattr(self, key, journey_dict[key])
 */

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_journey_dict = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_journey_dict,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_journey_dict)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 718, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_journey_dict = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 718, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric___init__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), __pyx_v_journey_dict);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric___init__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_journey_dict) {
  PyObject *__pyx_v_key = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "gtfspy/routing/label.pyx":719
 * 
 *     def __init__(self, journey_dict):
 *         for key in journey_dict:             # <<<<<<<<<<<<<<
 *             setattr(self, key, journey_dict[key])
 *         # Assert that key attributes are present
 */
  if (likely(PyList_CheckExact(__pyx_v_journey_dict)) || PyTuple_CheckExact(__pyx_v_journey_dict)) {
    __pyx_t_1 = __pyx_v_journey_dict; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_journey_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 719, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 719, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 719, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 719, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 719, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "gtfspy/routing/label.pyx":720
 *     def __init__(self, journey_dict):
 *         for key in journey_dict:
 *             setattr(self, key, journey_dict[key])             # <<<<<<<<<<<<<<
 *         # Assert that key attributes are present
 *         assert hasattr(self, "journey_id")
 */
    __pyx_t_4 = PyObject_GetItem(__pyx_v_journey_dict, __pyx_v_key); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_SetAttr(((PyObject *)__pyx_v_self), __pyx_v_key, __pyx_t_4); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "gtfspy/routing/label.pyx":719
 * 
 *     def __init__(self, journey_dict):
 *         for key in journey_dict:             # <<<<<<<<<<<<<<
 *             setattr(self, key, journey_dict[key])
 *         # Assert that key attributes are present
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":722
 *             setattr(self, key, journey_dict[key])
 *         # Assert that key attributes are present
 *         assert hasattr(self, "journey_id")             # <<<<<<<<<<<<<<
 *         assert hasattr(self, "from_stop_I")
 *         assert hasattr(self, "to_stop_I")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = PyObject_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_journey_id); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 722, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":723
 *         # Assert that key attributes are present
 *         assert hasattr(self, "journey_id")
 *         assert hasattr(self, "from_stop_I")             # <<<<<<<<<<<<<<
 *         assert hasattr(self, "to_stop_I")
 *         assert hasattr(self, "departure_time")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = PyObject_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_from_stop_I); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 723, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 723, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":724
 *         assert hasattr(self, "journey_id")
 *         assert hasattr(self, "from_stop_I")
 *         assert hasattr(self, "to_stop_I")             # <<<<<<<<<<<<<<
 *         assert hasattr(self, "departure_time")
 *         assert hasattr(self, "arrival_time_target")
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = PyObject_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_to_stop_I); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 724, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 724, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":725
 *         assert hasattr(self, "from_stop_I")
 *         assert hasattr(self, "to_stop_I")
 *         assert hasattr(self, "departure_time")             # <<<<<<<<<<<<<<
 *         assert hasattr(self, "arrival_time_target")
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = PyObject_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_departure_time); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 725, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 725, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":726
 *         assert hasattr(self, "to_stop_I")
 *         assert hasattr(self, "departure_time")
 *         assert hasattr(self, "arrival_time_target")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_6 = PyObject_HasAttr(((PyObject *)__pyx_v_self), __pyx_n_s_arrival_time_target); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 726, __pyx_L1_error)
    if (unlikely(!(__pyx_t_6 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 726, __pyx_L1_error)
    }
  }
  #endif

  /* "gtfspy/routing/label.pyx":718
 *     cdef public double departure_time, arrival_time_target
 * 
 *     def __init__(self, journey_dict):             # <<<<<<<<<<<<<<
 *         for key in journey_dict:
 *             setattr(self, key, journey_dict[key])
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":729
 * 
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_3__getstate__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_2__getstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_2__getstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__getstate__", 0);

  /* "gtfspy/routing/label.pyx":730
 * 
 *     def __getstate__(self):
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label             # <<<<<<<<<<<<<<
 * 
 *     def __setstate__(self, state):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_connection); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_previous_label); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":729
 * 
 * 
 *     def __getstate__(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__getstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":732
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_5__setstate__(PyObject *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_4__setstate__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_4__setstate__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__setstate__", 0);

  /* "gtfspy/routing/label.pyx":733
 * 
 *     def __setstate__(self, state):
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state             # <<<<<<<<<<<<<<
 * 
 *     def _tuple_for_ordering(self):
 */
  if ((likely(PyTuple_CheckExact(__pyx_v_state))) || (PyList_CheckExact(__pyx_v_state))) {
    PyObject* sequence = __pyx_v_state;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 5)) {
      if (size > 5) __Pyx_RaiseTooManyValuesError(5);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 733, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 4); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 4); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 5; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 733, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[5] = {&__pyx_t_1,&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_v_state); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
    for (index=0; index < 5; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 5) < 0) __PYX_ERR(0, 733, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 733, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->departure_time = __pyx_t_8;
  __pyx_v_self->arrival_time_target = __pyx_t_9;
  __pyx_v_self->movement_duration = __pyx_t_10;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_connection, __pyx_t_4) < 0) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_previous_label, __pyx_t_5) < 0) __PYX_ERR(0, 733, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "gtfspy/routing/label.pyx":732
 *         return self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label
 * 
 *     def __setstate__(self, state):             # <<<<<<<<<<<<<<
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__setstate__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":735
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.movement_duration
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_7_tuple_for_ordering(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_tuple_for_ordering (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_6_tuple_for_ordering(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_6_tuple_for_ordering(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_tuple_for_ordering", 0);

  /* "gtfspy/routing/label.pyx":736
 * 
 *     def _tuple_for_ordering(self):
 *         return self.departure_time, -self.arrival_time_target, -self.movement_duration             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(LabelGeneric self, LabelGeneric other, int op):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble((-__pyx_v_self->arrival_time_target)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int((-__pyx_v_self->movement_duration)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":735
 *         self.departure_time, self.arrival_time_target, self.movement_duration, self.connection, self.previous_label = state
 * 
 *     def _tuple_for_ordering(self):             # <<<<<<<<<<<<<<
 *         return self.departure_time, -self.arrival_time_target, -self.movement_duration
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric._tuple_for_ordering", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":738
 *         return self.departure_time, -self.arrival_time_target, -self.movement_duration
 * 
 *     def __richcmp__(LabelGeneric self, LabelGeneric other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "self", 0))) __PYX_ERR(0, 738, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 738, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_8__richcmp__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_8__richcmp__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_v_op) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "gtfspy/routing/label.pyx":739
 * 
 *     def __richcmp__(LabelGeneric self, LabelGeneric other, int op):
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":740
 *     def __richcmp__(LabelGeneric self, LabelGeneric other, int op):
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":741
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  __pyx_t_4 = ((__pyx_v_op == 2) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":742
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple             # <<<<<<<<<<<<<<
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 742, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":741
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if op == 2:  # ==             # <<<<<<<<<<<<<<
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 */
  }

  /* "gtfspy/routing/label.pyx":743
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  __pyx_t_4 = ((__pyx_v_op == 3) != 0);
  if (__pyx_t_4) {

    /* "gtfspy/routing/label.pyx":744
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple             # <<<<<<<<<<<<<<
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":743
 *         if op == 2:  # ==
 *             return self_tuple == other_tuple
 *         if op == 3:  # !=             # <<<<<<<<<<<<<<
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 */
  }

  /* "gtfspy/routing/label.pyx":745
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
  switch (__pyx_v_op) {
    case 0:

    /* "gtfspy/routing/label.pyx":746
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":745
 *         if op == 3:  # !=
 *             return self_tuple != other_tuple
 *         if op == 0:  # less than             # <<<<<<<<<<<<<<
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 */
    break;

    /* "gtfspy/routing/label.pyx":747
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    case 4:

    /* "gtfspy/routing/label.pyx":748
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":747
 *         if op == 0:  # less than
 *             return self_tuple < other_tuple
 *         elif op == 4:  # greater than             # <<<<<<<<<<<<<<
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 */
    break;

    /* "gtfspy/routing/label.pyx":749
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    case 1:

    /* "gtfspy/routing/label.pyx":750
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple             # <<<<<<<<<<<<<<
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 750, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":749
 *         elif op == 4:  # greater than
 *             return self_tuple > other_tuple
 *         elif op == 1:  # <=             # <<<<<<<<<<<<<<
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 */
    break;

    /* "gtfspy/routing/label.pyx":751
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    case 5:

    /* "gtfspy/routing/label.pyx":752
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=
 *             return self_tuple >= other_tuple             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates(self, LabelGeneric other):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyObject_RichCompare(__pyx_v_self_tuple, __pyx_v_other_tuple, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 752, __pyx_L1_error)
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":751
 *         elif op == 1:  # <=
 *             return self_tuple <= other_tuple
 *         elif op == 5:  # >=             # <<<<<<<<<<<<<<
 *             return self_tuple >= other_tuple
 * 
 */
    break;
    default: break;
  }

  /* "gtfspy/routing/label.pyx":738
 *         return self.departure_time, -self.arrival_time_target, -self.movement_duration
 * 
 *     def __richcmp__(LabelGeneric self, LabelGeneric other, int op):             # <<<<<<<<<<<<<<
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":754
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_tuple = NULL;
  PyObject *__pyx_v_other_tuple = NULL;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("dominates", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11dominates)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 754, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 754, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":765
 *             True if this ParetoTuple dominates the other, otherwise False
 *         """
 *         self_tuple = self._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":766
 *         """
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()             # <<<<<<<<<<<<<<
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_other), __pyx_n_s_tuple_for_ordering); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_other_tuple = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":767
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__17, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__18, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 767, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 767, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 767, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 767, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 767, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 767, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_4 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 767, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 767, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 767, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_any, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":768
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False             # <<<<<<<<<<<<<<
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":767
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  }

  /* "gtfspy/routing/label.pyx":769
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_self_tuple, 0, -1L, NULL, NULL, &__pyx_slice__19, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_other_tuple, 0, -1L, NULL, NULL, &__pyx_slice__20, 0, 1, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 769, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 769, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 769, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 769, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
      PyObject* sequence = __pyx_t_3;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 769, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_9 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 769, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_9)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_2 = __pyx_t_10(__pyx_t_9); if (unlikely(!__pyx_t_2)) goto __pyx_L12_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_9), 2) < 0) __PYX_ERR(0, 769, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L13_unpacking_done;
      __pyx_L12_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 769, __pyx_L1_error)
      __pyx_L13_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_s, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_o, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_s, __pyx_v_o, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_all, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_12) {
  } else {
    __pyx_t_11 = __pyx_t_12;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_other_tuple, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_11 = __pyx_t_12;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_11) {

    /* "gtfspy/routing/label.pyx":770
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 *             return False             # <<<<<<<<<<<<<<
 *         else:
 *             return True
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "gtfspy/routing/label.pyx":769
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  }

  /* "gtfspy/routing/label.pyx":772
 *             return False
 *         else:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelGeneric other):
 */
  /*else*/ {
    __pyx_r = 1;
    goto __pyx_L0;
  }

  /* "gtfspy/routing/label.pyx":754
 *             return self_tuple >= other_tuple
 * 
 *     cpdef int dominates(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         """
 *         Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_tuple);
  __Pyx_XDECREF(__pyx_v_other_tuple);
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_6gtfspy_7routing_5label_12LabelGeneric_10dominates[] = "\n        Compute whether this LabelWithNumberVehicles dominates the other LabelWithNumberVehicles\n        Parameters\n        ----------\n        other: LabelGeneric\n        Returns\n        -------\n        dominates: bint\n            True if this ParetoTuple dominates the other, otherwise False\n        ";
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11dominates(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 754, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10dominates(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10dominates(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.dominates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":774
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_dominates = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_fina); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_13dominates_ignoring_dep_time_finalization)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 774, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":776
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelGeneric other):
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dominates = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "gtfspy/routing/label.pyx":778
 *             self.arrival_time_target <= other.arrival_time_target
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelGeneric other):
 */
  __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_dominates); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 778, __pyx_L1_error)
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":774
 *             return True
 * 
 *     cpdef int dominates_ignoring_dep_time_finalization(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dominates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_13dominates_ignoring_dep_time_finalization(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 774, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_12dominates_ignoring_dep_time_finalization(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_12dominates_ignoring_dep_time_finalization(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_finalization", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_finalization(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time_finalization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":780
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_15dominates_ignoring_dep_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 780, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 780, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":784
 *             int dominates
 *         dominates = (
 *             (self.arrival_time_target <= other.arrival_time_target) or             # <<<<<<<<<<<<<<
 *             (self.arrival_time_target == other.arrival_time_target and self.movement_duration <= other.movement_duration)
 * 
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);
  if (!__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }

  /* "gtfspy/routing/label.pyx":785
 *         dominates = (
 *             (self.arrival_time_target <= other.arrival_time_target) or
 *             (self.arrival_time_target == other.arrival_time_target and self.movement_duration <= other.movement_duration)             # <<<<<<<<<<<<<<
 * 
 *         )
 */
  __pyx_t_7 = (__pyx_v_self->arrival_time_target == __pyx_v_other->arrival_time_target);
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_self->movement_duration <= __pyx_v_other->movement_duration);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L3_bool_binop_done:;
  __pyx_v_dominates = __pyx_t_6;

  /* "gtfspy/routing/label.pyx":788
 * 
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_time(self, LabelGeneric other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":780
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_15dominates_ignoring_dep_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 780, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14dominates_ignoring_dep_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14dominates_ignoring_dep_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":790
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17dominates_ignoring_time)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 790, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":794
 *             int dominates
 *         dominates = (
 *             self.movement_duration <= other.movement_duration             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_v_dominates = (__pyx_v_self->movement_duration <= __pyx_v_other->movement_duration);

  /* "gtfspy/routing/label.pyx":796
 *             self.movement_duration <= other.movement_duration
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelGeneric other):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":790
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_time(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17dominates_ignoring_time(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_time (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 790, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16dominates_ignoring_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16dominates_ignoring_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_time(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.dominates_ignoring_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":798
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static int __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other, int __pyx_skip_dispatch) {
  int __pyx_v_dominates;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dominates_ignoring_dep_time_and); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19dominates_ignoring_dep_time_and_n_boardings)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_other)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_other)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_other));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_other));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_other));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":802
 *             int dominates
 *         dominates = (
 *             self.arrival_time_target <= other.arrival_time_target             # <<<<<<<<<<<<<<
 *         )
 *         return dominates
 */
  __pyx_v_dominates = (__pyx_v_self->arrival_time_target <= __pyx_v_other->arrival_time_target);

  /* "gtfspy/routing/label.pyx":804
 *             self.arrival_time_target <= other.arrival_time_target
 *         )
 *         return dominates             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 */
  __pyx_r = __pyx_v_dominates;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":798
 *         return dominates
 * 
 *     cpdef int dominates_ignoring_dep_time_and_n_boardings(self, LabelGeneric other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int dominates
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19dominates_ignoring_dep_time_and_n_boardings(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, 1, "other", 0))) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_18dominates_ignoring_dep_time_and_n_boardings(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_18dominates_ignoring_dep_time_and_n_boardings(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dominates_ignoring_dep_time_and_n_boardings", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_and_n_boardings(__pyx_v_self, __pyx_v_other, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.dominates_ignoring_dep_time_and_n_boardings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":806
 *         return dominates
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_label_with_connection_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_21get_label_with_connection_added)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_connection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_connection};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 806, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_connection);
          __Pyx_GIVEREF(__pyx_v_connection);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_connection);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 806, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":807
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":808
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":807
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":808
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 */
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 808, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 808, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":807
 * 
 *     cpdef get_label_with_connection_added(self, connection):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric), __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":806
 *         return dominates
 * 
 *     cpdef get_label_with_connection_added(self, connection):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_21get_label_with_connection_added(PyObject *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_label_with_connection_added (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_20get_label_with_connection_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_connection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_20get_label_with_connection_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_label_with_connection_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_label_with_connection_added(__pyx_v_self, __pyx_v_connection, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_label_with_connection_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":809
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_23get_copy)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 809, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":810
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":811
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_connection); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":810
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":811
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_previous_label); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_previous_label, __pyx_t_3) < 0) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":810
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric), __pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":809
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, connection=connection, previous_label=self)
 *     cpdef get_copy(self):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_23get_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22get_copy(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22get_copy(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":813
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_departure_time, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_specified_departur); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_25get_copy_with_specified_departure_time)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_departure_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_departure_time};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 813, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_departure_time);
          __Pyx_GIVEREF(__pyx_v_departure_time);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_departure_time);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 813, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":814
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelGeneric(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":815
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelGeneric(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_connection); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "gtfspy/routing/label.pyx":814
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelGeneric(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "gtfspy/routing/label.pyx":815
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelGeneric(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)             # <<<<<<<<<<<<<<
 * 
 *     cpdef double duration(self):
 */
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_previous_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_previous_label, __pyx_t_2) < 0) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":814
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):
 *         return LabelGeneric(departure_time, self.arrival_time_target,             # <<<<<<<<<<<<<<
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric), __pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":813
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef get_copy_with_specified_departure_time(self, departure_time):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, self.arrival_time_target,
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_25get_copy_with_specified_departure_time(PyObject *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_24get_copy_with_specified_departure_time(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_departure_time));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_24get_copy_with_specified_departure_time(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_departure_time) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_specified_departure_time", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_specified_departure_time(__pyx_v_self, __pyx_v_departure_time, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy_with_specified_departure_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":817
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_6gtfspy_7routing_5label_12LabelGeneric_duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  __Pyx_RefNannySetupContext("duration", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_27duration)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 817, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 817, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":818
 * 
 *     cpdef double duration(self):
 *         return self.arrival_time_target - self.departure_time             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_r = (__pyx_v_self->arrival_time_target - __pyx_v_self->departure_time);
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":817
 *                                            self.movement_duration, self.connection, previous_label=self.previous_label)
 * 
 *     cpdef double duration(self):             # <<<<<<<<<<<<<<
 *         return self.arrival_time_target - self.departure_time
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("gtfspy.routing.label.LabelGeneric.duration", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_27duration(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("duration (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_26duration(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_26duration(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_duration(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":821
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label = {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_departure_time = 0;
  PyObject *__pyx_v_walk_duration = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("direct_walk_label (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_departure_time,&__pyx_n_s_walk_duration,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_departure_time)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, 1); __PYX_ERR(0, 821, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "direct_walk_label") < 0)) __PYX_ERR(0, 821, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_departure_time = values[0];
    __pyx_v_walk_duration = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("direct_walk_label", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 821, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_28direct_walk_label(__pyx_v_departure_time, __pyx_v_walk_duration);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_28direct_walk_label(PyObject *__pyx_v_departure_time, PyObject *__pyx_v_walk_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("direct_walk_label", 0);

  /* "gtfspy/routing/label.pyx":822
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)             # <<<<<<<<<<<<<<
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_departure_time, __pyx_v_walk_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_v_departure_time);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_departure_time);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_int_0);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_2, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":821
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.direct_walk_label", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":824
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection, int __pyx_skip_dispatch) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_copy_with_walk_added); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_31get_copy_with_walk_added)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_walk_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 824, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_connection};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_connection);
        __Pyx_GIVEREF(__pyx_v_connection);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_connection);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric))))) __PYX_ERR(0, 824, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "gtfspy/routing/label.pyx":825
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelGeneric(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->departure_time - __pyx_v_walk_duration)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "gtfspy/routing/label.pyx":826
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelGeneric(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->movement_duration + __pyx_v_walk_duration)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "gtfspy/routing/label.pyx":825
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelGeneric(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __Pyx_INCREF(Py_True);
  __Pyx_GIVEREF(Py_True);
  PyTuple_SET_ITEM(__pyx_t_7, 3, Py_True);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_4 = 0;

  /* "gtfspy/routing/label.pyx":826
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelGeneric(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_connection, __pyx_v_connection) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_previous_label, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 826, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":825
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):
 *         return LabelGeneric(self.departure_time - walk_duration,             # <<<<<<<<<<<<<<
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric), __pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":824
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 *     cpdef LabelGeneric get_copy_with_walk_added(self, double walk_duration, object connection):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(self.departure_time - walk_duration,
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_31get_copy_with_walk_added(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  double __pyx_v_walk_duration;
  PyObject *__pyx_v_connection = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_copy_with_walk_added (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_walk_duration,&__pyx_n_s_connection,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_walk_duration)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connection)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, 1); __PYX_ERR(0, 824, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_copy_with_walk_added") < 0)) __PYX_ERR(0, 824, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_walk_duration = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_walk_duration == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 824, __pyx_L3_error)
    __pyx_v_connection = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_copy_with_walk_added", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 824, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_30get_copy_with_walk_added(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), __pyx_v_walk_duration, __pyx_v_connection);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_30get_copy_with_walk_added(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, double __pyx_v_walk_duration, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_copy_with_walk_added", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_walk_added(__pyx_v_self, __pyx_v_walk_duration, __pyx_v_connection, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.get_copy_with_walk_added", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":828
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_33__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_33__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_32__str__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_32__str__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "gtfspy/routing/label.pyx":829
 * 
 *     def __str__(self):
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration))             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "gtfspy/routing/label.pyx":828
 *                                            self.arrival_time_target, self.movement_duration+walk_duration, True, connection=connection, previous_label=self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str((self.departure_time, self.arrival_time_target, self.movement_duration))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":714
 * cdef class LabelGeneric:
 *     # This class is used only for the analysis stage, when the data has been stored in a database
 *     cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \             # <<<<<<<<<<<<<<
 *         in_vehicle_duration, transfer_wait_duration, walking_duration, pre_journey_wait_fp
 *     cdef public double departure_time, arrival_time_target
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->journey_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.journey_id.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->journey_id = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.journey_id.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->from_stop_I); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.from_stop_I.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->from_stop_I = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.from_stop_I.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->to_stop_I); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.to_stop_I.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->to_stop_I = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.to_stop_I.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->n_boardings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.n_boardings.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->n_boardings = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.n_boardings.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->movement_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.movement_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->movement_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.movement_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->journey_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.journey_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_v_self->journey_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.journey_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":715
 *     # This class is used only for the analysis stage, when the data has been stored in a database
 *     cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \
 *         in_vehicle_duration, transfer_wait_duration, walking_duration, pre_journey_wait_fp             # <<<<<<<<<<<<<<
 *     cdef public double departure_time, arrival_time_target
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->in_vehicle_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.in_vehicle_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_v_self->in_vehicle_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.in_vehicle_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->transfer_wait_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.transfer_wait_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_v_self->transfer_wait_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.transfer_wait_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->walking_duration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.walking_duration.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_v_self->walking_duration = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.walking_duration.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->pre_journey_wait_fp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.pre_journey_wait_fp.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_v_self->pre_journey_wait_fp = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.pre_journey_wait_fp.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "gtfspy/routing/label.pyx":716
 *     cdef public int journey_id,  from_stop_I, to_stop_I, n_boardings, movement_duration, journey_duration, \
 *         in_vehicle_duration, transfer_wait_duration, walking_duration, pre_journey_wait_fp
 *     cdef public double departure_time, arrival_time_target             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, journey_dict):
 */

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->departure_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.departure_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L1_error)
  __pyx_v_self->departure_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.departure_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target___get__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target___get__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->arrival_time_target); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 716, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.arrival_time_target.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_2__set__(((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_2__set__(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L1_error)
  __pyx_v_self->arrival_time_target = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("gtfspy.routing.label.LabelGeneric.arrival_time_target.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeSimple __pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeSimple(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeSimple;
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeSimple(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_15LabelTimeSimple_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_15LabelTimeSimple_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_15LabelTimeSimple_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_15LabelTimeSimple_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelTimeSimple[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_5__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_7__setstate__, METH_O, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_9dominates, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_11dominates_ignoring_dep_time, METH_O, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_15duration, METH_NOARGS, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_17get_copy, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelTimeSimple[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_15LabelTimeSimple_departure_time, __pyx_setprop_6gtfspy_7routing_5label_15LabelTimeSimple_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_15LabelTimeSimple_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_15LabelTimeSimple_arrival_time_target, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelTimeSimple = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelTimeSimple", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeSimple, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    LabelTime describes entries in a Profile.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_3__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelTimeSimple, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelTimeSimple, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_15LabelTimeSimple_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelTimeSimple, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTime __pyx_vtable_6gtfspy_7routing_5label_LabelTime;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTime(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelTime;
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTime(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_9LabelTime_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_9LabelTime_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_9LabelTime_first_leg_is_walk(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_first_leg_is_walk(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_9LabelTime_17first_leg_is_walk_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelTime[] = {
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_3_tuple_for_ordering, METH_NOARGS, 0},
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_7__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_9__setstate__, METH_O, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_11dominates, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_13dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_15dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_17get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_19get_copy_with_specified_departure_time, METH_O, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_23duration, METH_NOARGS, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_9LabelTime_25get_copy_with_walk_added, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelTime[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_9LabelTime_departure_time, __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_9LabelTime_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_arrival_time_target, (char *)0, 0},
  {(char *)"first_leg_is_walk", __pyx_getprop_6gtfspy_7routing_5label_9LabelTime_first_leg_is_walk, __pyx_setprop_6gtfspy_7routing_5label_9LabelTime_first_leg_is_walk, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelTime = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelTime", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTime, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    LabelTime describes entries in a Profile.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_9LabelTime_5__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelTime, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelTime, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_9LabelTime_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelTime, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_n_boardings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_n_boardings(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11n_boardings_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_first_leg_is_walk(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_first_leg_is_walk(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17first_leg_is_walk_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_5__setstate__, METH_O, 0},
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_7_tuple_for_ordering, METH_NOARGS, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_11dominates, METH_O, __pyx_doc_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_10dominates},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_13dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_15dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_17dominates_ignoring_time, METH_O, 0},
  {"dominates_ignoring_dep_time_and_n_boardings", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_19dominates_ignoring_dep_time_and_n_boardings, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_21get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_23get_copy_with_specified_departure_time, METH_O, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_25duration, METH_NOARGS, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_29get_copy_with_walk_added, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_departure_time, __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_arrival_time_target, (char *)0, 0},
  {(char *)"n_boardings", __pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_n_boardings, __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_n_boardings, (char *)0, 0},
  {(char *)"first_leg_is_walk", __pyx_getprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_first_leg_is_walk, __pyx_setprop_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_first_leg_is_walk, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelTimeWithBoardingsCount", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_31__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelVehLegCount __pyx_vtable_6gtfspy_7routing_5label_LabelVehLegCount;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelVehLegCount(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelVehLegCount;
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelVehLegCount(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_n_boardings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_n_boardings(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11n_boardings_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_first_leg_is_walk(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_first_leg_is_walk(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17first_leg_is_walk_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelVehLegCount[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_5__setstate__, METH_O, 0},
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_7_tuple_for_ordering, METH_NOARGS, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_11dominates, METH_O, __pyx_doc_6gtfspy_7routing_5label_16LabelVehLegCount_10dominates},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_13dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_15dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_17get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_19get_copy_with_specified_departure_time, METH_O, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_21get_copy_with_walk_added, METH_O, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelVehLegCount[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_departure_time, __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_departure_time, (char *)0, 0},
  {(char *)"n_boardings", __pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_n_boardings, __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_n_boardings, (char *)0, 0},
  {(char *)"first_leg_is_walk", __pyx_getprop_6gtfspy_7routing_5label_16LabelVehLegCount_first_leg_is_walk, __pyx_setprop_6gtfspy_7routing_5label_16LabelVehLegCount_first_leg_is_walk, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelVehLegCount = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelVehLegCount", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelVehLegCount, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelVehLegCount, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelVehLegCount, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_16LabelVehLegCount_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelVehLegCount, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;
  p->previous_label = Py_None; Py_INCREF(Py_None);
  p->connection = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute(PyObject *o) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->previous_label);
  Py_CLEAR(p->connection);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)o;
  if (p->previous_label) {
    e = (*v)(p->previous_label, a); if (e) return e;
  }
  if (p->connection) {
    e = (*v)(p->connection, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *)o;
  tmp = ((PyObject*)p->previous_label);
  p->previous_label = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->connection);
  p->connection = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_n_boardings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_n_boardings(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11n_boardings_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_movement_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_movement_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17movement_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_first_leg_is_walk(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_first_leg_is_walk(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17first_leg_is_walk_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_previous_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_previous_label(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_14previous_label_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_connection(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_connection(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10connection_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_5__setstate__, METH_O, 0},
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_7_tuple_for_ordering, METH_NOARGS, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_11dominates, METH_O, __pyx_doc_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_10dominates},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_13dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_15dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_17dominates_ignoring_time, METH_O, 0},
  {"dominates_ignoring_dep_time_and_n_boardings", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_19dominates_ignoring_dep_time_and_n_boardings, METH_O, 0},
  {"get_label_with_connection_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_21get_label_with_connection_added, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_23get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_25get_copy_with_specified_departure_time, METH_O, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_27duration, METH_NOARGS, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_31get_copy_with_walk_added, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_departure_time, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_arrival_time_target, (char *)0, 0},
  {(char *)"n_boardings", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_n_boardings, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_n_boardings, (char *)0, 0},
  {(char *)"movement_duration", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_movement_duration, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_movement_duration, (char *)0, 0},
  {(char *)"first_leg_is_walk", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_first_leg_is_walk, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_first_leg_is_walk, (char *)0, 0},
  {(char *)"previous_label", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_previous_label, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_previous_label, (char *)0, 0},
  {(char *)"connection", __pyx_getprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_connection, __pyx_setprop_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_connection, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelTimeBoardingsAndRoute", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_33__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_traverse*/
  __pyx_tp_clear_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelTimeAndRoute __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelTimeAndRoute(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeAndRoute;
  p->previous_label = Py_None; Py_INCREF(Py_None);
  p->connection = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeAndRoute(PyObject *o) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->previous_label);
  Py_CLEAR(p->connection);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_6gtfspy_7routing_5label_LabelTimeAndRoute(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)o;
  if (p->previous_label) {
    e = (*v)(p->previous_label, a); if (e) return e;
  }
  if (p->connection) {
    e = (*v)(p->connection, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6gtfspy_7routing_5label_LabelTimeAndRoute(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *p = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *)o;
  tmp = ((PyObject*)p->previous_label);
  p->previous_label = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->connection);
  p->connection = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_movement_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_movement_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17movement_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_first_leg_is_walk(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_first_leg_is_walk(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17first_leg_is_walk_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_previous_label(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_previous_label(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_14previous_label_5__del__(o);
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_connection(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_connection(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_3__set__(o, v);
  }
  else {
    return __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_10connection_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelTimeAndRoute[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_5__setstate__, METH_O, 0},
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_7_tuple_for_ordering, METH_NOARGS, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_11dominates, METH_O, __pyx_doc_6gtfspy_7routing_5label_17LabelTimeAndRoute_10dominates},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_13dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_15dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_17dominates_ignoring_time, METH_O, 0},
  {"dominates_ignoring_dep_time_and_n_boardings", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_19dominates_ignoring_dep_time_and_n_boardings, METH_O, 0},
  {"get_label_with_connection_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_21get_label_with_connection_added, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_23get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_25get_copy_with_specified_departure_time, METH_O, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_27duration, METH_NOARGS, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_31get_copy_with_walk_added, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelTimeAndRoute[] = {
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_departure_time, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_arrival_time_target, (char *)0, 0},
  {(char *)"movement_duration", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_movement_duration, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_movement_duration, (char *)0, 0},
  {(char *)"first_leg_is_walk", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_first_leg_is_walk, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_first_leg_is_walk, (char *)0, 0},
  {(char *)"previous_label", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_previous_label, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_previous_label, (char *)0, 0},
  {(char *)"connection", __pyx_getprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_connection, __pyx_setprop_6gtfspy_7routing_5label_17LabelTimeAndRoute_connection, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelTimeAndRoute", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_33__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_traverse*/
  __pyx_tp_clear_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_17LabelTimeAndRoute_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelTimeAndRoute, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_6gtfspy_7routing_5label_LabelGeneric __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric;

static PyObject *__pyx_tp_new_6gtfspy_7routing_5label_LabelGeneric(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *)o);
  p->__pyx_vtab = __pyx_vtabptr_6gtfspy_7routing_5label_LabelGeneric;
  return o;
}

static void __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelGeneric(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_journey_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_journey_id(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_10journey_id_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_from_stop_I(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_from_stop_I(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11from_stop_I_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_to_stop_I(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_to_stop_I(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9to_stop_I_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_n_boardings(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_n_boardings(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11n_boardings_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_movement_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_movement_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17movement_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_journey_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_journey_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16journey_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_in_vehicle_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_in_vehicle_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19in_vehicle_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_transfer_wait_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_transfer_wait_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_22transfer_wait_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_walking_duration(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_walking_duration(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_16walking_duration_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_pre_journey_wait_fp(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_pre_journey_wait_fp(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19pre_journey_wait_fp_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_departure_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_departure_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_14departure_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_arrival_time_target(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_1__get__(o);
}

static int __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_arrival_time_target(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19arrival_time_target_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_6gtfspy_7routing_5label_LabelGeneric[] = {
  {"__getstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_3__getstate__, METH_NOARGS, 0},
  {"__setstate__", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_5__setstate__, METH_O, 0},
  {"_tuple_for_ordering", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_7_tuple_for_ordering, METH_NOARGS, 0},
  {"dominates", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_11dominates, METH_O, __pyx_doc_6gtfspy_7routing_5label_12LabelGeneric_10dominates},
  {"dominates_ignoring_dep_time_finalization", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_13dominates_ignoring_dep_time_finalization, METH_O, 0},
  {"dominates_ignoring_dep_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_15dominates_ignoring_dep_time, METH_O, 0},
  {"dominates_ignoring_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_17dominates_ignoring_time, METH_O, 0},
  {"dominates_ignoring_dep_time_and_n_boardings", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_19dominates_ignoring_dep_time_and_n_boardings, METH_O, 0},
  {"get_label_with_connection_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_21get_label_with_connection_added, METH_O, 0},
  {"get_copy", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_23get_copy, METH_NOARGS, 0},
  {"get_copy_with_specified_departure_time", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_25get_copy_with_specified_departure_time, METH_O, 0},
  {"duration", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_27duration, METH_NOARGS, 0},
  {"direct_walk_label", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_copy_with_walk_added", (PyCFunction)__pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_31get_copy_with_walk_added, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_6gtfspy_7routing_5label_LabelGeneric[] = {
  {(char *)"journey_id", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_journey_id, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_journey_id, (char *)0, 0},
  {(char *)"from_stop_I", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_from_stop_I, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_from_stop_I, (char *)0, 0},
  {(char *)"to_stop_I", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_to_stop_I, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_to_stop_I, (char *)0, 0},
  {(char *)"n_boardings", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_n_boardings, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_n_boardings, (char *)0, 0},
  {(char *)"movement_duration", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_movement_duration, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_movement_duration, (char *)0, 0},
  {(char *)"journey_duration", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_journey_duration, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_journey_duration, (char *)0, 0},
  {(char *)"in_vehicle_duration", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_in_vehicle_duration, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_in_vehicle_duration, (char *)0, 0},
  {(char *)"transfer_wait_duration", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_transfer_wait_duration, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_transfer_wait_duration, (char *)0, 0},
  {(char *)"walking_duration", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_walking_duration, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_walking_duration, (char *)0, 0},
  {(char *)"pre_journey_wait_fp", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_pre_journey_wait_fp, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_pre_journey_wait_fp, (char *)0, 0},
  {(char *)"departure_time", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_departure_time, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_departure_time, (char *)0, 0},
  {(char *)"arrival_time_target", __pyx_getprop_6gtfspy_7routing_5label_12LabelGeneric_arrival_time_target, __pyx_setprop_6gtfspy_7routing_5label_12LabelGeneric_arrival_time_target, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_6gtfspy_7routing_5label_LabelGeneric = {
  PyVarObject_HEAD_INIT(0, 0)
  "gtfspy.routing.label.LabelGeneric", /*tp_name*/
  sizeof(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6gtfspy_7routing_5label_LabelGeneric, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_33__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_9__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_6gtfspy_7routing_5label_LabelGeneric, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_6gtfspy_7routing_5label_LabelGeneric, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_6gtfspy_7routing_5label_12LabelGeneric_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6gtfspy_7routing_5label_LabelGeneric, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "label",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_k_Users_rmkujala_work_gtfspy_gtfs, sizeof(__pyx_k_Users_rmkujala_work_gtfspy_gtfs), 0, 0, 1, 0},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_any, __pyx_k_any, sizeof(__pyx_k_any), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arrival_time_target, __pyx_k_arrival_time_target, sizeof(__pyx_k_arrival_time_target), 0, 0, 1, 1},
  {&__pyx_n_s_best_label, __pyx_k_best_label, sizeof(__pyx_k_best_label), 0, 0, 1, 1},
  {&__pyx_n_s_candidate, __pyx_k_candidate, sizeof(__pyx_k_candidate), 0, 0, 1, 1},
  {&__pyx_n_s_candidate_is_dominated, __pyx_k_candidate_is_dominated, sizeof(__pyx_k_candidate_is_dominated), 0, 0, 1, 1},
  {&__pyx_n_s_candidates, __pyx_k_candidates, sizeof(__pyx_k_candidates), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_compute_pareto_front, __pyx_k_compute_pareto_front, sizeof(__pyx_k_compute_pareto_front), 0, 0, 1, 1},
  {&__pyx_n_s_compute_pareto_front_naive, __pyx_k_compute_pareto_front_naive, sizeof(__pyx_k_compute_pareto_front_naive), 0, 0, 1, 1},
  {&__pyx_n_s_compute_pareto_front_smart, __pyx_k_compute_pareto_front_smart, sizeof(__pyx_k_compute_pareto_front_smart), 0, 0, 1, 1},
  {&__pyx_n_s_connection, __pyx_k_connection, sizeof(__pyx_k_connection), 0, 0, 1, 1},
  {&__pyx_n_s_current_best_labels_wo_deptime, __pyx_k_current_best_labels_wo_deptime, sizeof(__pyx_k_current_best_labels_wo_deptime), 0, 0, 1, 1},
  {&__pyx_n_s_departure_time, __pyx_k_departure_time, sizeof(__pyx_k_departure_time), 0, 0, 1, 1},
  {&__pyx_n_s_direct_walk_label, __pyx_k_direct_walk_label, sizeof(__pyx_k_direct_walk_label), 0, 0, 1, 1},
  {&__pyx_n_s_dominated, __pyx_k_dominated, sizeof(__pyx_k_dominated), 0, 0, 1, 1},
  {&__pyx_n_s_dominates, __pyx_k_dominates, sizeof(__pyx_k_dominates), 0, 0, 1, 1},
  {&__pyx_n_s_dominates_ignoring_dep_time, __pyx_k_dominates_ignoring_dep_time, sizeof(__pyx_k_dominates_ignoring_dep_time), 0, 0, 1, 1},
  {&__pyx_n_s_dominates_ignoring_dep_time_and, __pyx_k_dominates_ignoring_dep_time_and, sizeof(__pyx_k_dominates_ignoring_dep_time_and), 0, 0, 1, 1},
  {&__pyx_n_s_dominates_ignoring_dep_time_fina, __pyx_k_dominates_ignoring_dep_time_fina, sizeof(__pyx_k_dominates_ignoring_dep_time_fina), 0, 0, 1, 1},
  {&__pyx_n_s_dominates_ignoring_time, __pyx_k_dominates_ignoring_time, sizeof(__pyx_k_dominates_ignoring_time), 0, 0, 1, 1},
  {&__pyx_n_s_dominator, __pyx_k_dominator, sizeof(__pyx_k_dominator), 0, 0, 1, 1},
  {&__pyx_n_s_duration, __pyx_k_duration, sizeof(__pyx_k_duration), 0, 0, 1, 1},
  {&__pyx_n_s_finalization, __pyx_k_finalization, sizeof(__pyx_k_finalization), 0, 0, 1, 1},
  {&__pyx_n_s_first_leg_is_walk, __pyx_k_first_leg_is_walk, sizeof(__pyx_k_first_leg_is_walk), 0, 0, 1, 1},
  {&__pyx_n_s_from_stop_I, __pyx_k_from_stop_I, sizeof(__pyx_k_from_stop_I), 0, 0, 1, 1},
  {&__pyx_n_s_get_copy, __pyx_k_get_copy, sizeof(__pyx_k_get_copy), 0, 0, 1, 1},
  {&__pyx_n_s_get_copy_with_specified_departur, __pyx_k_get_copy_with_specified_departur, sizeof(__pyx_k_get_copy_with_specified_departur), 0, 0, 1, 1},
  {&__pyx_n_s_get_copy_with_walk_added, __pyx_k_get_copy_with_walk_added, sizeof(__pyx_k_get_copy_with_walk_added), 0, 0, 1, 1},
  {&__pyx_n_s_get_label_with_connection_added, __pyx_k_get_label_with_connection_added, sizeof(__pyx_k_get_label_with_connection_added), 0, 0, 1, 1},
  {&__pyx_n_s_get_non_dominated_entries, __pyx_k_get_non_dominated_entries, sizeof(__pyx_k_get_non_dominated_entries), 0, 0, 1, 1},
  {&__pyx_n_s_gtfspy_routing_label, __pyx_k_gtfspy_routing_label, sizeof(__pyx_k_gtfspy_routing_label), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_n_boardings, __pyx_k_ignore_n_boardings, sizeof(__pyx_k_ignore_n_boardings), 0, 0, 1, 1},
  {&__pyx_n_s_inf, __pyx_k_inf, sizeof(__pyx_k_inf), 0, 0, 1, 1},
  {&__pyx_n_s_is_dominated, __pyx_k_is_dominated, sizeof(__pyx_k_is_dominated), 0, 0, 1, 1},
  {&__pyx_n_s_journey_dict, __pyx_k_journey_dict, sizeof(__pyx_k_journey_dict), 0, 0, 1, 1},
  {&__pyx_n_s_journey_id, __pyx_k_journey_id, sizeof(__pyx_k_journey_id), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_label, __pyx_k_label, sizeof(__pyx_k_label), 0, 0, 1, 1},
  {&__pyx_n_s_label_list, __pyx_k_label_list, sizeof(__pyx_k_label_list), 0, 0, 1, 1},
  {&__pyx_n_s_labels, __pyx_k_labels, sizeof(__pyx_k_labels), 0, 0, 1, 1},
  {&__pyx_n_s_labels_other, __pyx_k_labels_other, sizeof(__pyx_k_labels_other), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_merge_pareto_frontiers, __pyx_k_merge_pareto_frontiers, sizeof(__pyx_k_merge_pareto_frontiers), 0, 0, 1, 1},
  {&__pyx_n_s_merge_pareto_frontiers_locals__g, __pyx_k_merge_pareto_frontiers_locals__g, sizeof(__pyx_k_merge_pareto_frontiers_locals__g), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_min_arrival_time_target, __pyx_k_min_arrival_time_target, sizeof(__pyx_k_min_arrival_time_target), 0, 0, 1, 1},
  {&__pyx_n_s_min_arrival_time_target_locals_l, __pyx_k_min_arrival_time_target_locals_l, sizeof(__pyx_k_min_arrival_time_target_locals_l), 0, 0, 1, 1},
  {&__pyx_n_s_min_n_vehicle_trips, __pyx_k_min_n_vehicle_trips, sizeof(__pyx_k_min_n_vehicle_trips), 0, 0, 1, 1},
  {&__pyx_n_s_min_n_vehicle_trips_locals_lambd, __pyx_k_min_n_vehicle_trips_locals_lambd, sizeof(__pyx_k_min_n_vehicle_trips_locals_lambd), 0, 0, 1, 1},
  {&__pyx_n_s_movement_duration, __pyx_k_movement_duration, sizeof(__pyx_k_movement_duration), 0, 0, 1, 1},
  {&__pyx_n_s_n_boardings, __pyx_k_n_boardings, sizeof(__pyx_k_n_boardings), 0, 0, 1, 1},
  {&__pyx_n_s_n_vehicle_trips, __pyx_k_n_vehicle_trips, sizeof(__pyx_k_n_vehicle_trips), 0, 0, 1, 1},
  {&__pyx_n_s_new_best, __pyx_k_new_best, sizeof(__pyx_k_new_best), 0, 0, 1, 1},
  {&__pyx_n_s_new_label, __pyx_k_new_label, sizeof(__pyx_k_new_label), 0, 0, 1, 1},
  {&__pyx_n_s_new_remaining, __pyx_k_new_remaining, sizeof(__pyx_k_new_remaining), 0, 0, 1, 1},
  {&__pyx_n_s_old_partial_best, __pyx_k_old_partial_best, sizeof(__pyx_k_old_partial_best), 0, 0, 1, 1},
  {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
  {&__pyx_n_s_other_i, __pyx_k_other_i, sizeof(__pyx_k_other_i), 0, 0, 1, 1},
  {&__pyx_n_s_pareto_front, __pyx_k_pareto_front, sizeof(__pyx_k_pareto_front), 0, 0, 1, 1},
  {&__pyx_n_s_possible_dominators, __pyx_k_possible_dominators, sizeof(__pyx_k_possible_dominators), 0, 0, 1, 1},
  {&__pyx_n_s_previous_label, __pyx_k_previous_label, sizeof(__pyx_k_previous_label), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_remaining, __pyx_k_remaining, sizeof(__pyx_k_remaining), 0, 0, 1, 1},
  {&__pyx_n_s_reversed, __pyx_k_reversed, sizeof(__pyx_k_reversed), 0, 0, 1, 1},
  {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
  {&__pyx_n_s_survived, __pyx_k_survived, sizeof(__pyx_k_survived), 0, 0, 1, 1},
  {&__pyx_n_s_survivor_list, __pyx_k_survivor_list, sizeof(__pyx_k_survivor_list), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_to_stop_I, __pyx_k_to_stop_I, sizeof(__pyx_k_to_stop_I), 0, 0, 1, 1},
  {&__pyx_n_s_tuple_for_ordering, __pyx_k_tuple_for_ordering, sizeof(__pyx_k_tuple_for_ordering), 0, 0, 1, 1},
  {&__pyx_n_s_walk_duration, __pyx_k_walk_duration, sizeof(__pyx_k_walk_duration), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_builtin_all = __Pyx_GetBuiltinName(__pyx_n_s_all); if (!__pyx_builtin_all) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_builtin_any = __Pyx_GetBuiltinName(__pyx_n_s_any); if (!__pyx_builtin_any) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_builtin_reversed = __Pyx_GetBuiltinName(__pyx_n_s_reversed); if (!__pyx_builtin_reversed) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 389, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 440, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "gtfspy/routing/label.pyx":420
 *     pareto_front_merged: list[LabelTime]
 *     """
 *     def _get_non_dominated_entries(candidates, possible_dominators, survivor_list=None):             # <<<<<<<<<<<<<<
 *         if survivor_list is None:
 *             survivor_list = list()
 */
  __pyx_tuple__6 = PyTuple_Pack(6, __pyx_n_s_candidates, __pyx_n_s_possible_dominators, __pyx_n_s_survivor_list, __pyx_n_s_candidate, __pyx_n_s_candidate_is_dominated, __pyx_n_s_dominator); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(3, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_get_non_dominated_entries, 420, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 420, __pyx_L1_error)
  __pyx_tuple__8 = PyTuple_Pack(1, ((PyObject *)Py_None)); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "gtfspy/routing/label.pyx":511
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_slice__9 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__9)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__9);
  __Pyx_GIVEREF(__pyx_slice__9);
  __pyx_slice__10 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__10)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__10);
  __Pyx_GIVEREF(__pyx_slice__10);

  /* "gtfspy/routing/label.pyx":513
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);
  __pyx_slice__12 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__12)) __PYX_ERR(0, 513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__12);
  __Pyx_GIVEREF(__pyx_slice__12);

  /* "gtfspy/routing/label.pyx":643
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_slice__13 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__13)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__13);
  __Pyx_GIVEREF(__pyx_slice__13);
  __pyx_slice__14 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__14)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__14);
  __Pyx_GIVEREF(__pyx_slice__14);

  /* "gtfspy/routing/label.pyx":645
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_slice__15 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__15)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__15);
  __Pyx_GIVEREF(__pyx_slice__15);
  __pyx_slice__16 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__16)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__16);
  __Pyx_GIVEREF(__pyx_slice__16);

  /* "gtfspy/routing/label.pyx":767
 *         self_tuple = self._tuple_for_ordering()
 *         other_tuple = other._tuple_for_ordering()
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):             # <<<<<<<<<<<<<<
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:
 */
  __pyx_slice__17 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__17)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__17);
  __Pyx_GIVEREF(__pyx_slice__17);
  __pyx_slice__18 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__18)) __PYX_ERR(0, 767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__18);
  __Pyx_GIVEREF(__pyx_slice__18);

  /* "gtfspy/routing/label.pyx":769
 *         if any([(s < o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]):
 *             return False
 *         elif all([(s == o) for s, o in zip(self_tuple[:-1], other_tuple[:-1])]) and self_tuple[-1] < other_tuple[-1]:             # <<<<<<<<<<<<<<
 *             return False
 *         else:
 */
  __pyx_slice__19 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__19)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__19);
  __Pyx_GIVEREF(__pyx_slice__19);
  __pyx_slice__20 = PySlice_New(Py_None, __pyx_int_neg_1, Py_None); if (unlikely(!__pyx_slice__20)) __PYX_ERR(0, 769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__20);
  __Pyx_GIVEREF(__pyx_slice__20);

  /* "gtfspy/routing/label.pyx":43
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 43, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":111
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 111, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 111, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":224
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 224, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":303
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(0, departure_time, True)
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 303, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 303, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":314
 * #     return first.dominates(other)
 * 
 * def compute_pareto_front_smart(list label_list):             # <<<<<<<<<<<<<<
 *     return compute_pareto_front(label_list)
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_label_list); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_compute_pareto_front_smart, 314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 314, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":317
 *     return compute_pareto_front(label_list)
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):             # <<<<<<<<<<<<<<
 *     pareto_front = []
 *     if len(label_list) == 0:
 */
  __pyx_tuple__31 = PyTuple_Pack(12, __pyx_n_s_label_list, __pyx_n_s_finalization, __pyx_n_s_ignore_n_boardings, __pyx_n_s_pareto_front, __pyx_n_s_label, __pyx_n_s_dominates, __pyx_n_s_current_best_labels_wo_deptime, __pyx_n_s_new_label, __pyx_n_s_is_dominated, __pyx_n_s_best_label, __pyx_n_s_new_best, __pyx_n_s_old_partial_best); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 12, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_compute_pareto_front, 317, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 317, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":357
 *     return pareto_front
 * 
 * def compute_pareto_front_naive(list label_list):             # <<<<<<<<<<<<<<
 *     """
 *     Computes the Pareto frontier of a given label_list
 */
  __pyx_tuple__33 = PyTuple_Pack(9, __pyx_n_s_label_list, __pyx_n_s_dominated, __pyx_n_s_pareto_front, __pyx_n_s_remaining, __pyx_n_s_candidate, __pyx_n_s_new_remaining, __pyx_n_s_is_dominated, __pyx_n_s_other_i, __pyx_n_s_other); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 9, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_compute_pareto_front_naive, 357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":407
 *     return pareto_front
 * 
 * def merge_pareto_frontiers(labels, labels_other):             # <<<<<<<<<<<<<<
 *     """
 *     Merge two pareto frontiers by removing dominated entries.
 */
  __pyx_tuple__35 = PyTuple_Pack(5, __pyx_n_s_labels, __pyx_n_s_labels_other, __pyx_n_s_get_non_dominated_entries, __pyx_n_s_get_non_dominated_entries, __pyx_n_s_survived); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_merge_pareto_frontiers, 407, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":438
 *     return survived
 * 
 * def min_arrival_time_target(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_n_s_label_list); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_min_arrival_time_target, 438, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 438, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":444
 *         return float('inf')
 * 
 * def min_n_vehicle_trips(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_n_s_label_list); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_min_n_vehicle_trips, 444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 444, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":574
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_tuple__41 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 574, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 574, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":701
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 701, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 701, __pyx_L1_error)

  /* "gtfspy/routing/label.pyx":821
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(2, __pyx_n_s_departure_time, __pyx_n_s_walk_duration); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_rmkujala_work_gtfspy_gtfs, __pyx_n_s_direct_walk_label, 821, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initlabel(void); /*proto*/
PyMODINIT_FUNC initlabel(void)
#else
PyMODINIT_FUNC PyInit_label(void); /*proto*/
PyMODINIT_FUNC PyInit_label(void)
#endif
{
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_label(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("label", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_gtfspy__routing__label) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "gtfspy.routing.label")) {
      if (unlikely(PyDict_SetItemString(modules, "gtfspy.routing.label", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeSimple = &__pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeSimple.get_copy = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeSimple *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_15LabelTimeSimple_get_copy;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelTimeSimple) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelTimeSimple.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelTimeSimple.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeSimple) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelTimeSimple", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelTimeSimple) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple = &__pyx_type_6gtfspy_7routing_5label_LabelTimeSimple;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelTime = &__pyx_vtable_6gtfspy_7routing_5label_LabelTime;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime._tuple_for_ordering = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime__tuple_for_ordering;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_dominates_ignoring_dep_time_finalization;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.get_copy = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.get_copy_with_specified_departure_time = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, double, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_specified_departure_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTime.get_copy_with_walk_added = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTime *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_9LabelTime_get_copy_with_walk_added;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelTime) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelTime.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelTime.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelTime) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelTime", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelTime) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelTime = &__pyx_type_6gtfspy_7routing_5label_LabelTime;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount = &__pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_finalization;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.dominates_ignoring_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.dominates_ignoring_dep_time_and_n_boardings = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_dominates_ignoring_dep_time_and_n_boardings;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.get_copy = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.get_copy_with_specified_departure_time = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_specified_departure_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.get_copy_with_walk_added = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount *, double, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_get_copy_with_walk_added;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelTimeWithBoardingsCount", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount = &__pyx_type_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelVehLegCount = &__pyx_vtable_6gtfspy_7routing_5label_LabelVehLegCount;
  __pyx_vtable_6gtfspy_7routing_5label_LabelVehLegCount.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelVehLegCount.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelVehLegCount.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, struct __pyx_obj_6gtfspy_7routing_5label_LabelVehLegCount *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_16LabelVehLegCount_dominates_ignoring_dep_time_finalization;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelVehLegCount) < 0) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelVehLegCount.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelVehLegCount.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelVehLegCount) < 0) __PYX_ERR(0, 235, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelVehLegCount", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelVehLegCount) < 0) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount = &__pyx_type_6gtfspy_7routing_5label_LabelVehLegCount;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute = &__pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_finalization;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.dominates_ignoring_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.dominates_ignoring_dep_time_and_n_boardings = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_dominates_ignoring_dep_time_and_n_boardings;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.get_label_with_connection_added = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_label_with_connection_added;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.get_copy = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.get_copy_with_specified_departure_time = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_specified_departure_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.get_copy_with_walk_added = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute *, double, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_get_copy_with_walk_added;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelTimeBoardingsAndRoute", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute) < 0) __PYX_ERR(0, 450, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute = &__pyx_type_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeAndRoute = &__pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_finalization;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.dominates_ignoring_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.dominates_ignoring_dep_time_and_n_boardings = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_dominates_ignoring_dep_time_and_n_boardings;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.get_label_with_connection_added = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_label_with_connection_added;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.get_copy = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.get_copy_with_specified_departure_time = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_specified_departure_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelTimeAndRoute.get_copy_with_walk_added = (struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelTimeAndRoute *, double, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_17LabelTimeAndRoute_get_copy_with_walk_added;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelTimeAndRoute) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelTimeAndRoute", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute = &__pyx_type_6gtfspy_7routing_5label_LabelTimeAndRoute;
  __pyx_vtabptr_6gtfspy_7routing_5label_LabelGeneric = &__pyx_vtable_6gtfspy_7routing_5label_LabelGeneric;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.dominates = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.dominates_ignoring_dep_time_finalization = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_finalization;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.dominates_ignoring_dep_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.dominates_ignoring_time = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.dominates_ignoring_dep_time_and_n_boardings = (int (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_dominates_ignoring_dep_time_and_n_boardings;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.get_label_with_connection_added = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_label_with_connection_added;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.get_copy = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.get_copy_with_specified_departure_time = (PyObject *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_specified_departure_time;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.duration = (double (*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_duration;
  __pyx_vtable_6gtfspy_7routing_5label_LabelGeneric.get_copy_with_walk_added = (struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *(*)(struct __pyx_obj_6gtfspy_7routing_5label_LabelGeneric *, double, PyObject *, int __pyx_skip_dispatch))__pyx_f_6gtfspy_7routing_5label_12LabelGeneric_get_copy_with_walk_added;
  if (PyType_Ready(&__pyx_type_6gtfspy_7routing_5label_LabelGeneric) < 0) __PYX_ERR(0, 712, __pyx_L1_error)
  __pyx_type_6gtfspy_7routing_5label_LabelGeneric.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_6gtfspy_7routing_5label_LabelGeneric.tp_dict, __pyx_vtabptr_6gtfspy_7routing_5label_LabelGeneric) < 0) __PYX_ERR(0, 712, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "LabelGeneric", (PyObject *)&__pyx_type_6gtfspy_7routing_5label_LabelGeneric) < 0) __PYX_ERR(0, 712, __pyx_L1_error)
  __pyx_ptype_6gtfspy_7routing_5label_LabelGeneric = &__pyx_type_6gtfspy_7routing_5label_LabelGeneric;
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "gtfspy/routing/label.pyx":9
 *         public double arrival_time_target
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf')):             # <<<<<<<<<<<<<<
 *         self.departure_time = departure_time
 *         self.arrival_time_target = arrival_time_target
 */
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 9, __pyx_L1_error)
  __pyx_k_ = (-__pyx_t_1);
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 9, __pyx_L1_error)
  __pyx_k__2 = __pyx_t_1;

  /* "gtfspy/routing/label.pyx":43
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_15LabelTimeSimple_13direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":42
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple);

  /* "gtfspy/routing/label.pyx":43
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":42
 *         return self.arrival_time_target <= other.arrival_time_target
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTimeSimple(departure_time, departure_time + walk_duration)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeSimple);

  /* "gtfspy/routing/label.pyx":61
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, double departure_time=-float("inf"), double arrival_time_target=float('inf'), bint first_leg_is_walk=False,             # <<<<<<<<<<<<<<
 *                  **kwargs):
 *         self.departure_time = departure_time
 */
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_k__3 = (-__pyx_t_1);
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
  __pyx_k__4 = __pyx_t_1;

  /* "gtfspy/routing/label.pyx":111
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_9LabelTime_21direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":110
 *         return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTime);

  /* "gtfspy/routing/label.pyx":111
 * 
 *     @staticmethod
 *     def direct_walk_label(double departure_time, double walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":110
 *         return LabelTime(departure_time, self.arrival_time_target, self.first_leg_is_walk)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(double departure_time, double walk_duration):
 *         return LabelTime(departure_time, departure_time + walk_duration, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTime->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTime);

  /* "gtfspy/routing/label.pyx":224
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_27LabelTimeWithBoardingsCount_27direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":223
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount);

  /* "gtfspy/routing/label.pyx":224
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":223
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeWithBoardingsCount(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeWithBoardingsCount);

  /* "gtfspy/routing/label.pyx":241
 *         public bint first_leg_is_walk
 * 
 *     def __init__(self, n_boardings=0, double departure_time=-float('inf'), first_leg_is_walk=False, **kwargs):             # <<<<<<<<<<<<<<
 *         self.n_boardings = n_boardings
 *         self.departure_time = departure_time
 */
  __pyx_t_1 = __Pyx_PyObject_AsDouble(__pyx_n_s_inf); if (unlikely(__pyx_t_1 == ((double)-1) && PyErr_Occurred())) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_k__5 = (-__pyx_t_1);

  /* "gtfspy/routing/label.pyx":303
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(0, departure_time, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_16LabelVehLegCount_23direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":302
 *                                 first_leg_is_walk=True)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelVehLegCount(0, departure_time, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount);

  /* "gtfspy/routing/label.pyx":303
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelVehLegCount(0, departure_time, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":302
 *                                 first_leg_is_walk=True)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelVehLegCount(0, departure_time, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelVehLegCount);

  /* "gtfspy/routing/label.pyx":314
 * #     return first.dominates(other)
 * 
 * def compute_pareto_front_smart(list label_list):             # <<<<<<<<<<<<<<
 *     return compute_pareto_front(label_list)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_1compute_pareto_front_smart, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compute_pareto_front_smart, __pyx_t_2) < 0) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":317
 *     return compute_pareto_front(label_list)
 * 
 * def compute_pareto_front(list label_list, finalization=False, ignore_n_boardings=False):             # <<<<<<<<<<<<<<
 *     pareto_front = []
 *     if len(label_list) == 0:
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_3compute_pareto_front, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compute_pareto_front, __pyx_t_2) < 0) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":357
 *     return pareto_front
 * 
 * def compute_pareto_front_naive(list label_list):             # <<<<<<<<<<<<<<
 *     """
 *     Computes the Pareto frontier of a given label_list
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_5compute_pareto_front_naive, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compute_pareto_front_naive, __pyx_t_2) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":407
 *     return pareto_front
 * 
 * def merge_pareto_frontiers(labels, labels_other):             # <<<<<<<<<<<<<<
 *     """
 *     Merge two pareto frontiers by removing dominated entries.
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_7merge_pareto_frontiers, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_merge_pareto_frontiers, __pyx_t_2) < 0) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":438
 *     return survived
 * 
 * def min_arrival_time_target(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.arrival_time_target).arrival_time_target
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_9min_arrival_time_target, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_min_arrival_time_target, __pyx_t_2) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":444
 *         return float('inf')
 * 
 * def min_n_vehicle_trips(label_list):             # <<<<<<<<<<<<<<
 *     if len(label_list) > 0:
 *         return min(label_list, key=lambda label: label.n_vehicle_trips).n_vehicle_trips
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_11min_n_vehicle_trips, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_min_n_vehicle_trips, __pyx_t_2) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "gtfspy/routing/label.pyx":574
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_26LabelTimeBoardingsAndRoute_29direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":573
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute);

  /* "gtfspy/routing/label.pyx":574
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":573
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeBoardingsAndRoute(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeBoardingsAndRoute);

  /* "gtfspy/routing/label.pyx":701
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_17LabelTimeAndRoute_29direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":700
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute);

  /* "gtfspy/routing/label.pyx":701
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":700
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelTimeAndRoute(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelTimeAndRoute);

  /* "gtfspy/routing/label.pyx":821
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6gtfspy_7routing_5label_12LabelGeneric_29direct_walk_label, NULL, __pyx_n_s_gtfspy_routing_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":820
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric);

  /* "gtfspy/routing/label.pyx":821
 * 
 *     @staticmethod
 *     def direct_walk_label(departure_time, walk_duration):             # <<<<<<<<<<<<<<
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 * 
 */
  __pyx_t_2 = __Pyx_GetNameInClass((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric, __pyx_n_s_direct_walk_label); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "gtfspy/routing/label.pyx":820
 *         return self.arrival_time_target - self.departure_time
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def direct_walk_label(departure_time, walk_duration):
 *         return LabelGeneric(departure_time, departure_time + walk_duration, 0, True)
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_staticmethod, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric->tp_dict, __pyx_n_s_direct_walk_label, __pyx_t_2) < 0) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_6gtfspy_7routing_5label_LabelGeneric);

  /* "gtfspy/routing/label.pyx":1
 * cdef class LabelTimeSimple:             # <<<<<<<<<<<<<<
 *     """
 *     LabelTime describes entries in a Profile.
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init gtfspy.routing.label", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init gtfspy.routing.label");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ExtTypeTest */
    static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* GetModuleGlobalName */
    static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* IterNext */
      static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
#if CYTHON_USE_TYPE_SLOTS
    if (unlikely(!iternext)) {
#else
    if (unlikely(!iternext) || unlikely(!PyIter_Check(iterator))) {
#endif
        PyErr_Format(PyExc_TypeError,
            "%.200s object is not an iterator", Py_TYPE(iterator)->tp_name);
        return NULL;
    }
    next = iternext(iterator);
    if (likely(next))
        return next;
#if CYTHON_USE_TYPE_SLOTS
#if PY_VERSION_HEX >= 0x02070000
    if (unlikely(iternext == &_PyObject_NextNotImplemented))
        return NULL;
#endif
#endif
    if (defval) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
            if (unlikely(exc_type != PyExc_StopIteration) &&
                    !PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))
                return NULL;
            PyErr_Clear();
        }
        Py_INCREF(defval);
        return defval;
    }
    if (!PyErr_Occurred())
        PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}

/* GetItemInt */
        static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallMethod1 */
        static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto done;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
done:
    Py_XDECREF(method);
    return result;
}

/* append */
        static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* FetchCommonType */
        static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
        static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* SliceObject */
            static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* SetVTable */
            static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* pyobject_as_double */
            static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
    float_value = PyNumber_Float(obj);  if (0) goto bad;
#else
    PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
    if (likely(nb) && likely(nb->nb_float)) {
        float_value = nb->nb_float(obj);
        if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
            PyErr_Format(PyExc_TypeError,
                "__float__ returned non-float (type %.200s)",
                Py_TYPE(float_value)->tp_name);
            Py_DECREF(float_value);
            goto bad;
        }
    } else if (PyUnicode_CheckExact(obj) || PyBytes_CheckExact(obj)) {
#if PY_MAJOR_VERSION >= 3
        float_value = PyFloat_FromString(obj);
#else
        float_value = PyFloat_FromString(obj, 0);
#endif
    } else {
        PyObject* args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        PyTuple_SET_ITEM(args, 0, obj);
        float_value = PyObject_Call((PyObject*)&PyFloat_Type, args, 0);
        PyTuple_SET_ITEM(args, 0, 0);
        Py_DECREF(args);
    }
#endif
    if (likely(float_value)) {
        double value = PyFloat_AS_DOUBLE(float_value);
        Py_DECREF(float_value);
        return value;
    }
bad:
    return (double)-1;
}

/* GetNameInClass */
            static PyObject *__Pyx_GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result)
        result = __Pyx_GetModuleGlobalName(name);
    return result;
}

/* CodeObjectCache */
            static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
            #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
            #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
            static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
            static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
            static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
